from hwt.hdlObjects.types.integer import Integer
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.hdlType import HdlType
from hwt.hdlObjects.types.defs import BOOL
from hwt.hdlObjects.types.booleanVal import BooleanVal


def convertSimInteger__val(self, sigOrVal, toType):
    if isinstance(toType, Integer):
        v = sigOrVal.clone()
        if toType.min is not None:
            assert v.val >= toType.min
        if toType.max is not None:
            assert v.val <= toType.max
                
        v._dtype = toType
        return v
    elif toType == BOOL:
        v = sigOrVal.val
        assert v == 0 or v == 1
        return BooleanVal(v, BOOL, sigOrVal.vldMask, sigOrVal.updateTime)
            
    elif isinstance(toType, Bits):
        _v = sigOrVal.val 
        v = toType.fromPy(_v & toType._allMask)
        v.updateTime = sigOrVal.updateTime
         
        v._dtype = toType
        if not sigOrVal.vldMask:
            v.vldMask = 0
        return v
            
    return HdlType.defaultConvert(self, sigOrVal, toType)



from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class SignalItem(object):
    """basic hdl signal"""
    def __init__(self, name, dtype, defaultVal=None,  virtualOnly=False):
        """
        @param name: name for better orientation in netlists (used only in serialization)
        @param dtype: data type of this signal
        @param defaultVal: value for initialization 
        @param virtualOnly: flag indicates that this assignments is only virtual and should not be added into
                netlist, because it is only for internal notation
        """
        self.name = name
        self._dtype = dtype
        self.virtualOnly = virtualOnly
        if defaultVal is None:
            defaultVal = dtype.fromPy(None) 
        self.defaultVal = defaultVal
        self._setDefValue()
        
    def _setDefValue(self):
        v = self.defaultVal
        if isinstance(v, RtlSignalBase):
            v = v.staticEval()
            
        self._val = v.clone()
        self._oldVal = self._val.clone()
        self._oldVal.vldMask = 0
    
    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer, onlyPrintDefaultValues
        return VhdlSerializer.SignalItem(self, onlyPrintDefaultValues)        
    
        
        



import difflib
from inspect import isgenerator
import os
import pprint
import unittest
from unittest.util import safe_repr, _common_shorten_repr

from hwt.hdlObjects.value import Value
from hwt.simulator.agentConnector import valToInt
from hwt.simulator.configVhdlTestbench import HdlSimConfigVhdlTestbench
from hwt.simulator.hdlSimulator import HdlSimulator
from hwt.simulator.shortcuts import simPrepare
from hwt.simulator.simSignal import SimSignal
from hwt.simulator.utils import agent_randomize
from hwt.simulator.vcdHdlSimConfig import VcdHdlSimConfig
from _random import Random
from hwt.hdlObjects.constants import Time
from unittest.case import _Outcome


def allValuesToInts(sequenceOrVal):
    if isinstance(sequenceOrVal, Value):
        return valToInt(sequenceOrVal)
    elif not sequenceOrVal:
        return sequenceOrVal
    elif isinstance(sequenceOrVal, (list, tuple)) or isgenerator(sequenceOrVal):
        l = []
        for i in sequenceOrVal:
            l.append(allValuesToInts(i))

        if isinstance(sequenceOrVal, tuple):
            return tuple(l)

        return l
    else:
        return sequenceOrVal


class SimTestCase(unittest.TestCase):
    """
    This is TestCase class contains methods which are usually used during
    hdl simulation.

    @attention: self.model, self.procs has to be specified before running doSim
    u = Axi_rDatapump()
    self.model, self.procs = simPrepare(u)
    """
    _defaultSeed = 317
    _rand = Random(_defaultSeed)

    def getTestName(self):
        className, testName = self.id().split(".")[-2:]
        return "%s_%s" % (className, testName)

    def doSim(self, time):
        outputFileName = "tmp/" + self.getTestName() + ".vcd"
        d = os.path.dirname(outputFileName)
        if d:
            os.makedirs(d, exist_ok=True)
        with open(outputFileName, 'w') as outputFile:
            # return _simUnitVcd(simModel, stimulFunctions, outputFile=f, time=time)
            sim = HdlSimulator()

            # configure simulator to log in vcd
            sim.config = VcdHdlSimConfig(outputFile)

            # run simulation, stimul processes are register after initial initialization
            sim.simUnit(self.model, time=time, extraProcesses=self.procs)
            return sim

    def dumpHdlTestbench(self, time, file=None):
        if file:
            outputFileName = file
        else:
            outputFileName = "tmp/" + self.getTestName() + "_tb.vhd"
        d = os.path.dirname(outputFileName)
        if d:
            os.makedirs(d, exist_ok=True)
        with open(outputFileName, 'w') as outputFile:
            # return _simUnitVcd(simModel, stimulFunctions, outputFile=f, time=time)
            sim = HdlSimulator()

            # configure simulator to log in vcd
            sim.config = HdlSimConfigVhdlTestbench(self.u)

            # run simulation, stimul processes are register after initial initialization
            sim.simUnit(self.model, time=time, extraProcesses=self.procs)

            sim.config.dump(outputFile)

            return sim

    def assertValEqual(self, first, second, msg=None):
        if isinstance(first, SimSignal):
            first = first._val
        if not isinstance(first, int):
            first = valToInt(first)

        return unittest.TestCase.assertEqual(self, first, second, msg=msg)

    def assertEmpty(self, val, msg=None):
        return unittest.TestCase.assertEqual(self, len(val), 0, msg=msg)

    def assertValSequenceEqual(self, seq1, seq2, msg=None, seq_type=None):
        """
        An equality assertion for ordered sequences (like lists and tuples).
        For the purposes of this function, a valid ordered sequence type is one
        which can be indexed, has a length, and has an equality operator.

        Args:
        @param seq1: can contain instance of values or nested list of them
        @param seq2: items are not converted
        @param seq_type: The expected datatype of the sequences, or None if no
                    datatype should be enforced.
        @param msg: Optional message to use on failure instead of a list of
                    differences.
        """
        seq1 = allValuesToInts(seq1)
        if seq_type is not None:
            seq_type_name = seq_type.__name__
            if not isinstance(seq1, seq_type):
                raise self.failureException('First sequence is not a %s: %s'
                                        % (seq_type_name, safe_repr(seq1)))
            if not isinstance(seq2, seq_type):
                raise self.failureException('Second sequence is not a %s: %s'
                                        % (seq_type_name, safe_repr(seq2)))
        else:
            seq_type_name = "sequence"

        differing = None
        try:
            len1 = len(seq1)
        except (TypeError, NotImplementedError):
            differing = 'First %s has no length.    Non-sequence?' % (
                    seq_type_name)

        if differing is None:
            try:
                len2 = len(seq2)
            except (TypeError, NotImplementedError):
                differing = 'Second %s has no length.    Non-sequence?' % (
                        seq_type_name)

        if differing is None:
            if seq1 == seq2:
                return

            differing = '%ss differ: %s != %s\n' % (
                    (seq_type_name.capitalize(),) + 
                    _common_shorten_repr(seq1, seq2))

            for i in range(min(len1, len2)):
                try:
                    item1 = seq1[i]
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('\nUnable to index element %d of first %s\n' % 
                                 (i, seq_type_name))
                    break

                try:
                    item2 = seq2[i]
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('\nUnable to index element %d of second %s\n' % 
                                 (i, seq_type_name))
                    break

                if item1 != item2:
                    differing += ('\nFirst differing element %d:\n%s\n%s\n' % 
                                 ((i,) + _common_shorten_repr(item1, item2)))
                    break
            else:
                if (len1 == len2 and seq_type is None and
                    type(seq1) != type(seq2)):
                    # The sequences are the same, but have differing types.
                    return

            if len1 > len2:
                differing += ('\nFirst %s contains %d additional '
                             'elements.\n' % (seq_type_name, len1 - len2))
                try:
                    differing += ('First extra element %d:\n%s\n' % 
                                  (len2, safe_repr(seq1[len2])))
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('Unable to index element %d '
                                  'of first %s\n' % (len2, seq_type_name))
            elif len1 < len2:
                differing += ('\nSecond %s contains %d additional '
                             'elements.\n' % (seq_type_name, len2 - len1))
                try:
                    differing += ('First extra element %d:\n%s\n' % 
                                  (len1, safe_repr(seq2[len1])))
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('Unable to index element %d '
                                  'of second %s\n' % (len1, seq_type_name))
        standardMsg = differing
        diffMsg = '\n' + '\n'.join(
            difflib.ndiff(pprint.pformat(seq1).splitlines(),
                          pprint.pformat(seq2).splitlines()))

        standardMsg = self._truncateMessage(standardMsg, diffMsg)
        msg = self._formatMessage(msg, standardMsg)
        self.fail(msg)

    def randomize(self, intf):
        self.procs.append(agent_randomize(intf._ag,
                                          50 * Time.ns,
                                          seed=self._rand.getrandbits(64)))

    def prepareUnit(self, u, modelCls=None, dumpModelIn=None, onAfterToRtl=None):
        self.u, self.model, self.procs = simPrepare(u,
                                                    modelCls=modelCls,
                                                    dumpModelIn=dumpModelIn,
                                                    onAfterToRtl=onAfterToRtl)
    def setUp(self):
        self._rand.seed(self._defaultSeed)



# byteplay: CPython assembler/disassembler
# Copyright (C) 2006 Noam Raphael | Version: http://code.google.com/p/byteplay
# Rewritten 2009 Demur Rumed | Version: http://github.com/serprex/byteplay
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
__version__ = '0.2'

from dis import findlabels
import opcode

from hwt.pyUtils.optCodes import _se, _rf, Opcode, hascode, LOAD_CONST, hasarg, \
     hasconst, hasname, hasjabs, hasflow, hasjump, haslocal, hascompare, \
     hasfree, hasjrel, cmp_op, SETUP_FINALLY, BREAK_LOOP, RETURN_VALUE, RAISE_VARARGS, \
    FOR_ITER, JUMP_FORWARD, JUMP_ABSOLUTE, JUMP_IF_FALSE_OR_POP, \
    JUMP_IF_TRUE_OR_POP, CONTINUE_LOOP, SETUP_LOOP, SETUP_EXCEPT, POP_BLOCK, \
    END_FINALLY, WITH_CLEANUP, STORE_NAME, LOAD_NAME, DELETE_NAME, YIELD_VALUE, \
    LOAD_GLOBAL, LOAD_FAST, CALL_FUNCTION, POP_TOP
from types import CodeType
import dis

def getse(op, arg=None):
    if op in _se:return _se[op]
    if arg is None:raise ValueError("%s requires arg" % op)
    if op in _rf:return _rf[op](arg)
    raise ValueError("Unknown %s %s" % (op, arg))

class Label(object):
    pass

class SetLinenoType(object):
    def __repr__(self):
        return 'SetLineno'
SetLineno = SetLinenoType()


def isopcode(x):
    return x is not SetLineno and not isinstance(x, Label)

CO_OPTIMIZED = 1
CO_NEWLOCALS = 2
CO_VARARGS = 4
CO_VARKEYWORDS = 8
CO_NESTED = 16
CO_GENERATOR = 32
CO_NOFREE = 64
CO_GENERATOR_ALLOWED = 0x1000
CO_FUTURE_DIVISION = 0x2000
CO_FUTURE_ABSOLUTE_IMPORT = 0x4000
CO_FUTURE_WITH_STATEMENT = 0x8000

class Code(object):
    """An object which holds all the information which a Python code object
    holds, but in an easy-to-play-with representation

    The attributes are:

    Affecting action
    code - list of 2-tuples: the code
    freevars - list of strings: the free vars of the code (those are names
               of variables created in outer functions and used in the function)
    args - list of strings: the arguments of the code
    varargs - boolean: Does args end with a '*args' argument
    varkwargs - boolean: Does args end with a '**kwargs' argument
    newlocals - boolean: Should a new local namespace be created.
                (True in functions, False for module and exec code)

    Not affecting action
    name - string: the name of the code (co_name)
    filename - string: the file name of the code (co_filename)
    firstlineno - int: the first line number (co_firstlineno)
    docstring - string or None: the docstring (the first item of co_consts,
                if it's str or unicode)

    code is a list of 2-tuples. The first item is an opcode, or SetLineno, or a
    Label instance. The second item is the argument, if applicable, or None"""
    def __init__(self, code, freevars, args, varargs, \
                  varkwargs, newlocals, name, filename, firstlineno, docstring):
        self.code = code
        self.freevars = freevars
        self.args = args
        self.varargs = varargs
        self.varkwargs = varkwargs
        self.newlocals = newlocals
        self.name = name
        self.filename = filename
        self.firstlineno = firstlineno
        self.docstring = docstring
        
    @staticmethod
    def _findlinestarts(code):
        """Find the offsets in a byte code which are start of lines in the source
        Generate pairs offset,lineno as described in Python/compile.c
        This is a modified version of dis.findlinestarts, which allows multiplelinestarts with the same line number"""
        lineno = code.co_firstlineno
        addr = 0
        for byte_incr, line_incr in zip(code.co_lnotab[0::2], code.co_lnotab[1::2]):
            if byte_incr:
                yield addr, lineno
                addr += byte_incr
            lineno += line_incr
        yield addr, lineno
        
    @classmethod
    def from_code(cls, co):
        """Disassemble a Python code object into a Code object"""
        co_code = co.co_code
        labels = dict((addr, Label()) for addr in findlabels(co_code))
        linestarts = dict(cls._findlinestarts(co))
        cellfree = co.co_cellvars + co.co_freevars
        code = []
        n = len(co_code)
        i = extended_arg = 0
        while i < n:
            op = Opcode(co_code[i])
            if i in labels:
                code.append((labels[i], None))
            if i in linestarts:
                code.append((SetLineno, linestarts[i]))
            i += 1
            if op in hascode:
                lastop, lastarg = code[-1]
                if lastop != LOAD_CONST:
                    raise ValueError("%s should be preceded by LOAD_CONST code" % op)
                code[-1] = (LOAD_CONST, Code.from_code(lastarg))
            if op not in hasarg:
                code.append((op, None))
            else:
                arg = co_code[i] | co_code[i + 1] << 8 | extended_arg
                extended_arg = 0
                i += 2
                if op == opcode.EXTENDED_ARG:extended_arg = arg << 16
                else:
                    code.append((op, co.co_consts[arg] if op in hasconst else co.co_names[arg] if op in hasname else labels[arg] if op in hasjabs else labels[i + arg] if op in hasjrel else co.co_varnames[arg] if op in haslocal else cmp_op[arg] if op in hascompare else cellfree[arg] if op in hasfree else arg))
        varargs = not not co.co_flags & CO_VARARGS
        varkwargs = not not co.co_flags & CO_VARKEYWORDS
        return cls(code=code,
                freevars=co.co_freevars,
                args=co.co_varnames[:co.co_argcount + varargs + varkwargs],
                varargs=varargs,
                varkwargs=varkwargs,
                newlocals=not not co.co_flags & CO_NEWLOCALS,
                name=co.co_name,
                filename=co.co_filename,
                firstlineno=co.co_firstlineno,
                docstring=co.co_consts[0] if co.co_consts and isinstance(co.co_consts[0], str) else None)
        
    def __eq__(self, other):
        try:
            if(self.freevars != other.freevars or
            self.args != other.args or
            self.varargs != other.varargs or
            self.varkwargs != other.varkwargs or
            self.newlocals != other.newlocals or
            self.name != other.name or
            self.filename != other.filename or
            self.firstlineno != other.firstlineno or
            self.docstring != other.docstring or
            len(self.code) != len(other.code)):return False
            # This isn't trivial due to labels
            lmap = {}
            for (op1, arg1), (op2, arg2) in zip(self.code, other.code):
                if isinstance(op1, Label):
                    if lmap.setdefault(arg1, arg2) is not arg2:return False
                else:
                    if op1 != op2:return False
                    if op1 in hasjump:
                        if lmap.setdefault(arg1, arg2) is not arg2:return False
                    elif arg1 != arg2:return False
            return True
        except:
            return False
        
    def _compute_stacksize(self):
        code = self.code
        label_pos = dict((op[0], pos) for pos, op in enumerate(code) if isinstance(op[0], Label))
        # sf_targets are the targets of SETUP_FINALLY opcodes. They are recorded
        # because they have special stack behaviour. If an exception was raised
        # in the block pushed by a SETUP_FINALLY opcode, the block is popped
        # and 3 objects are pushed. On return or continue, the block is popped
        # and 2 objects are pushed. If nothing happened, the block is popped by
        # a POP_BLOCK opcode and 1 object is pushed by a (LOAD_CONST, None)
        # operation
        # Our solution is to record the stack state of SETUP_FINALLY targets
        # as having 3 objects pushed, which is the maximum. However, to make
        # stack recording consistent, the get_next_stacks function will always
        # yield the stack state of the target as if 1 object was pushed, but
        # this will be corrected in the actual stack recording
        sf_targets = set(label_pos[arg] for op, arg in code if op == SETUP_FINALLY)
        stacks = [None] * len(code)
        maxsize = 0
        op = [(0, (0,))]
        def newstack(n):
            if curstack[-1] < -n:raise ValueError("Popped a non-existing element at %s %s" % (pos, code[pos - 3:pos + 2]))
            return curstack[:-1] + (curstack[-1] + n,)
        while op:
            pos, curstack = op.pop()
            o = sum(curstack)
            if o > maxsize:maxsize = o
            o, arg = code[pos]
            if isinstance(o, Label):
                if pos in sf_targets:curstack = curstack[:-1] + (curstack[-1] + 2,)
                if stacks[pos] is None:
                    stacks[pos] = curstack
                    if o not in (BREAK_LOOP, RETURN_VALUE, RAISE_VARARGS):
                        pos += 1
                        if not isopcode(o):
                            op += (pos, curstack),
                        elif o not in hasflow:
                            op += (pos, newstack(getse(o, arg))),
                        elif o == FOR_ITER:
                            op += (label_pos[arg], newstack(-1)), (pos, newstack(1))
                        elif o in (JUMP_FORWARD, JUMP_ABSOLUTE):
                            op += (label_pos[arg], curstack),
                        elif o in (JUMP_IF_FALSE_OR_POP, JUMP_IF_TRUE_OR_POP):
                            op += (label_pos[arg], curstack), (pos, curstack)
                        elif o == CONTINUE_LOOP:
                            op += (label_pos[arg], curstack[:-1]),
                        elif o == SETUP_LOOP:
                            op += (pos, curstack + (0,)), (label_pos[arg], curstack)
                        elif o == SETUP_EXCEPT:
                            op += (pos, curstack + (0,)), (label_pos[arg], newstack(3))
                        elif o == SETUP_FINALLY:
                            op += (pos, curstack + (0,)), (label_pos[arg], newstack(1))
                        elif o == POP_BLOCK:
                            op += (pos, curstack[:-1]),
                        elif o == END_FINALLY:
                            op += (pos, newstack(-3)),
                        elif o == WITH_CLEANUP:
                            op += (pos, newstack(-1)),
                        else:raise ValueError("Unhandled opcode %s" % op)
                elif stacks[pos] != curstack:
                    o = pos + 1
                    while code[o][0] not in hasflow:o += 1
                    if code[o][0] not in (RETURN_VALUE, RAISE_VARARGS):
                        raise ValueError("Inconsistent code at %s %s %s\n%s" % (pos, curstack, stacks[pos], code[pos - 5:pos + 4]))
        return maxsize
    
    def to_code(self):
        """Assemble a Python code object from a Code object"""
        co_argcount = len(self.args) - self.varargs - self.varkwargs
        co_stacksize = self._compute_stacksize()
        co_flags = set(op[0] for op in self.code)
        co_flags = (not(STORE_NAME in co_flags or LOAD_NAME in co_flags or DELETE_NAME in co_flags))  \
                    | (self.newlocals and CO_NEWLOCALS)                                               \
                    | (self.varargs and CO_VARARGS)                                                   \
                    | (self.varkwargs and CO_VARKEYWORDS)                                             \
                    | ((YIELD_VALUE in co_flags) << 5) | ((not co_flags & hasfree) << 6)
        co_consts = [self.docstring]
        co_names = []
        co_varnames = list(self.args)
        co_freevars = tuple(self.freevars)
        # Find all cellvars beforehand for two reasons
        # Need the number of them to construct the numeric arg for ops in hasfree
        # Need to put args which are cells in the beginning of co_cellvars
        cellvars = set(arg for op, arg in self.code
                if isopcode(op) and op in hasfree
                and arg not in co_freevars)
        co_cellvars = [jumps for jumps in self.args if jumps in cellvars]

        def index(seq, item, eq=True, can_append=True):
            for i, x in enumerate(seq):
                if x == item if eq else x is item:return i
            if can_append:
                seq.append(item)
                return len(seq) - 1
            else:
                raise IndexError("Item not found")

        jumps = []
        label_pos = {}
        lastlineno = self.firstlineno
        lastlinepos = 0
        co_code = bytearray()
        co_lnotab = bytearray()
        for i, (op, arg) in enumerate(self.code):
            if isinstance(op, Label):
                label_pos[op] = len(co_code)
            elif op is SetLineno:
                incr_lineno = arg - lastlineno
                incr_pos = len(co_code) - lastlinepos
                lastlineno = arg
                lastlinepos += incr_pos
                if not (incr_lineno or incr_pos):co_lnotab += b"\0\0"
                else:
                    while incr_pos > 255:
                        co_lnotab += b"\xFF\0"
                        incr_pos -= 255
                    while incr_lineno > 255:
                        co_lnotab += b"%c\xFF" % incr_pos
                        incr_pos = 0
                        incr_lineno -= 255
                    if incr_pos or incr_lineno:
                        co_lnotab += ("%c%c" % (incr_pos, incr_lineno)).encode()
            elif op == opcode.EXTENDED_ARG:
                self.code[i + 1][1] |= 1 << 32
            elif op not in hasarg:
                co_code += ("%c" % op).encode()
            else:
                if op in hasconst:
                    if isinstance(arg, Code) and i + 1 < len(self.code) and self.code[i + 1][0] in hascode:
                        arg = arg.to_code()
                    arg = index(co_consts, arg, 0)
                elif op in hasname:
                    arg = index(co_names, arg)
                elif op in hasjump:
                    jumps.append((len(co_code), arg))
                    arg = None
                elif op in haslocal:
                    arg = index(co_varnames, arg)
                elif op in hascompare:
                    arg = index(cmp_op, arg, can_append=False)
                elif op in hasfree:
                    try:
                        arg = index(co_freevars, arg, can_append=False) + len(cellvars)
                    except IndexError:
                        arg = index(co_cellvars, arg)
                if arg is not None and arg > 0xFFFF:
                    co_code += ("%c%c%c" % (opcode.EXTENDED_ARG, arg >> 16 & 0xFF, arg >> 24 & 0xFF)).encode()
                co_code += ("%c\0\0" % op if arg is None else "%c%c%c" % (op, arg & 0xFF, arg >> 8 & 0xFF)).encode()
        for pos, label in jumps:
            jump = label_pos[label]
            if co_code[pos] in hasjrel:
                jump -= pos + 3
            if jump > 0xFFFF:
                raise NotImplementedError("Extended jumps not implemented")
            co_code[pos + 1] = jump & 0xFF
            co_code[pos + 2] = jump >> 8
            
            
        if not self.varkwargs:
            kwonlyargcount = 0
        else:
            raise NotImplementedError()
            
        return CodeType(co_argcount, kwonlyargcount, len(co_varnames),
                        co_stacksize, co_flags, bytes(co_code),
                        tuple(co_consts), tuple(co_names), tuple(co_varnames),
                         self.filename, self.name, self.firstlineno,
                         bytes(co_lnotab), co_freevars, tuple(co_cellvars))
    
    def insertInCode(self, stratIndex, instructions):
        for i in instructions:
            c.code.insert(stratIndex, i)
            stratIndex += 1
            
    
    def addPositionalArg(self, argName):
        self.args = (argName,) + self.args

if __name__ == "__main__":
    def fn(a, a_en, b):
        print('With synthesizer')
        if a_en:
            return a
    
        
    c = Code.from_code(fn.__code__)
    argName = "__synthesizer__"
    c.addPositionalArg(argName)
    
    
    
    for i in c.code:
        print(i)
    
    #instr = [
    #    (Opcode(LOAD_GLOBAL), 'print'),
    #    (Opcode(LOAD_CONST), 'With synthesizer'),
    #    #(Opcode(LOAD_FAST), '__synthesizer__')  ,
    #    (Opcode(CALL_FUNCTION), 1),
    #    (Opcode(POP_TOP), None),
    #]
    #c.insertInCode(1, instr)
    print()
    for i in c.code:
        print(i)
    
    dis.dis(fn.__code__)
    
    fn.__code__ = c.to_code()
    print("\n\n")
    dis.dis(fn.__code__)
    print(fn(True, 10, True, 20))
    
    



from hwt.hdlObjects.value import Value, areValues
from hwt.hdlObjects.types.defs import BOOL
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps

BoolVal = BOOL.getValueCls()

class EnumVal(Value):
    @classmethod
    def fromPy(cls, val, typeObj):
        """
        @param val: value of python type bool or None
        @param typeObj: instance of HdlType
        """
        if val is None:
            valid = False
            val = typeObj._allValues[0]
        else:
            assert isinstance(val, str)
            valid = True
        
        return cls(val, typeObj, valid)
    
    def _eq__val(self, other):
        eq = self.val == other.val \
             and self.vldMask == other.vldMask == 1
        
        vldMask = int(self.vldMask == other.vldMask == 1)
        updateTime = max(self.updateTime, other.updateTime)
        return BoolVal(eq, BOOL, vldMask, updateTime)
    def _eq(self, other):
        assert self._dtype is other._dtype
        
        if areValues(self, other):
            return self._eq__val(other)
        else:
            return Operator.withRes(AllOps.EQ, [self, other], BOOL)
    
    
    def _ne__val(self, other):
        neq = self.val != other.val \
            and self.vldMask == other.vldMask == 1
        
        vldMask = int(self.vldMask == other.vldMask == 1)
        updateTime = max(self.updateTime, other.updateTime)
        return BoolVal(neq, BOOL, vldMask, updateTime)   
     
    def __ne__(self, other):
        assert self._dtype is other._dtype
        
        if areValues(self, other):
            return self._ne__val(other)
        else:
            return Operator.withRes(AllOps.NEQ, [self, other], BOOL)
        



from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.synthesizer.rtlLevel.signalUtils.walkers import discoverEventDependency

class DepContainer():
    def __init__(self):
        self.pos = set()  # if cond
        self.neg = set()  # if not cond
        
    def __repr__(self):
        return "<DepContainer pos:%s, neg:%s>" % (repr(self.pos), repr(self.neg))

def getBaseCond(c):
    """
    if is negated return original cond and negated flag
    """
    isNegated = False
    drivers = []
    try:
        drivers = c.drivers
    except AttributeError:
        pass
    if len(drivers) == 1:
        d = list(c.drivers)[0]
        if isinstance(d, Operator) and d.operator == AllOps.NOT:
            c = d.ops[0]
            isNegated = True
    return (c, isNegated)

    
def buildTermMapFromConditions(assignments):
    """
    Generate dict condition:DepContainer
    for each condition term in conditions of assignments
    """
    termMap = {}
    
    def insertToMap(condSig, assigment, isNegated):
        try:
            cont = termMap[condSig]
        except KeyError:
            cont = DepContainer()
            termMap[condSig] = cont
        
        if isNegated:
            c = cont.neg
        else:
            c = cont.pos
        c.add(assigment)
    
    def registerToMap(assigment):
        # walk all assignments and register them in term map
        # cond is set of term in conjunctive form
        for realC, isNegated in assigment._unresolvedConds:
            insertToMap(realC, assigment, isNegated)
            
    # resolve main hierarchy of conditions
    for a in assignments:
        registerToMap(a)
    
    return termMap

def countCondOccurrences(termMap):
    for cond, container in termMap.items():
        cnt = len(container.pos) + len(container.neg)
        maxStmId = 0
        for c in [container.pos, container.neg]:
            maxStmId = max([maxStmId, *map(lambda x: x._instId, c)]) 
        yield TermUsageRecord(cond, cnt, maxStmId)

def sortCondsByMostImpact(countedConds):
    for c in sorted(countedConds,
                    key=lambda x: (x.usedCnt, x.isEventDependent, x.maxStmId, x.term._instId),
                     reverse=True):
        yield c.term

class TermUsageRecord(object):
    def __init__(self, term, usedCnt, maxStmId):
        self.term = term
        self.usedCnt = usedCnt
        self.isEventDependent = any(discoverEventDependency(term))
        self.maxStmId = maxStmId
    
    
def extractCondTermOrder(assignments):
            # register assignments in tree of IfTreeNodes
    for a in assignments:
        # prepare base conds
        a._unresolvedConds = [ getBaseCond(c) for c in a.cond ]
    
    termMap = buildTermMapFromConditions(assignments)
    cntTerms = countCondOccurrences(termMap)
    yield from sortCondsByMostImpact(cntTerms)

class TermResolveReport():
    def __init__(self, usageCnt, maxStmId):
        self.usageCnt = usageCnt
        self.maxStmId = maxStmId

    

def extractCondTermOrderNonResolved(assignments, globalCondOrder):
    """
    Extract condition order from non resolved condition terms in assignments
    """
    termUsage = {}
    resolved = [] 
    
    for a in assignments:
        # add assignments without cond as statements
        if len(a._unresolvedConds) == 0:
            resolved.append(a)
        else:
            # else register this to termap to resolve most impact condition
            for c in a._unresolvedConds:
                try:
                    r = termUsage[c[0]]
                except KeyError:
                    r = TermResolveReport(0, 0)
                    termUsage[c[0]] = r
                finally:
                    r.usageCnt += 1
                    r.maxStmId = max(a._instId, r.maxStmId)

    def trrSortKey(term):
        termResRep = termUsage[term] 
        return (termResRep.usageCnt, globalCondOrder.index(term), termResRep.maxStmId)
                    
    # id is used to make sorting more deterministic, but it is not optimal solution
    condOrder = sorted(termUsage.keys(),
                      key=trrSortKey,
                      reverse=True)
    return resolved, condOrder


from hwt.hdlObjects.constants import INTF_DIRECTION
from hwt.synthesizer.interfaceLevel.unit import Unit
from hwt.synthesizer.exceptions import IntfLvlConfErr

def setOut(*interfaces, defVal=None):
    """
    set interfaces as output interfaces
    @attention: only for instances of EmptyUnit
    """
    for i in interfaces:
        i._setDirectionsLikeIn(INTF_DIRECTION.SLAVE)
        
class EmptyUnit(Unit):
    """
    Unit used for prototyping all output interfaces are connected to _defaultValue
    and this is only think which architecture contains
    
    @cvar _defaultValue: this value is used to initialize all signals 
    """
    _defaultValue = None
    def _toRtl(self):
        assert not self._wasSynthetised()
        if not hasattr(self, "_name"):
            self._name = self._getDefaultName()
        for i in self._interfaces:
            i._setDirectionsLikeIn(INTF_DIRECTION.MASTER)
        
        
        self._loadMyImplementations()
        # construct globals (generics for entity)
        self._cntx.globals = self._globalsFromParams()
        externInterf = [] 
        # prepare connections     
        for i in self._interfaces:
            signals = i._signalsForInterface(self._cntx)
            if not i._isExtern:
                raise IntfLvlConfErr("All interfaces in EmptyUnit has to be extern, %s: %s is not" % 
                                     (self.__class__.__name__, i._getFullName()))
            externInterf.extend(signals)
            # i._resolveDirections()
            # connect outputs to dummy value
            for s in signals:
                if s._interface._direction == INTF_DIRECTION.SLAVE:
                    s ** s._dtype.fromPy(self._defaultValue)
                    
        if not externInterf:
            raise  Exception("Can not find any external interface for unit " + self._name \
                              + "- there is no such a thing as unit without interfaces")
        yield from self._synthetiseContext(externInterf)
        # self._checkEntityPortDirections()



{{indent}}c_{{indentNum}}, cVld_{{ indentNum }} = simEvalCond(sim, {{ cond }})
{{indent}}#if ():
{{indent}}if c_{{indentNum}} or not cVld_{{ indentNum }}: 
{{indent}}    cVld_{{ indentNum }} = cVld_{{ indentNum-1 }} and cVld_{{ indentNum }}{%
if ifTrue|length > 0 %}{% 
    for stm in ifTrue %}
{{stm}}{% 
    endfor %}{% 
else %}
{{indent}}    #enclosure
{{enclosure}}{%
endif %}
{{indent}}#else:
{{indent}}if not c_{{indentNum}} or not cVld_{{ indentNum }}:{% 
if ifFalse|length > 0 %}{%
    for stm in ifFalse %}
{{stm}}{%
    endfor%}{% 
else %}
{{indent}}    #enclosure
{{enclosure}}
{% endif %}
{#
""" This template constructs if-else trees like this """ 
if cond:
    ...
else:
    ...
    
condDefined = True
cond = cond.eval()
# if cond:
if cond or cond == UNKNOWN:
    condDefined_0 = condDefined and cond != UNKNOWN  
    ...
# else:
if not cond or cond == UNKNOWN:
    condDefined_0 = condDefined_0 or cond != UNKNOWN
    ...
#}


from hwt.serializer.ip_packager.interfaces.intfConfig import IntfConfig, DEFAULT_CLOCK 
from hwt.interfaces.std import Rst, Rst_n
from hwt.pyUtils.arrayQuery import where


class IP_Clk(IntfConfig):
    def __init__(self):
        super().__init__()
        self.name = "clock"
        self.version = "1.0"
        self.vendor = "xilinx.com" 
        self.library = "signal"
        self.map = 'CLK'
            
    def postProcess(self, component, entity, allInterfaces, thisIf):
            rst = list(where(allInterfaces, lambda intf: isinstance(intf, Rst_n) 
                                                        or isinstance(intf, Rst)))
            if len(rst) > 0:
                rst = rst[0]
                self.addSimpleParam(thisIf, "ASSOCIATED_RESET", rst._name)  # getResetPortName
                
            elif len(rst) > 1:
                raise Exception("Don't know how to work with multiple resets")
            
            intfs = where(allInterfaces, lambda intf: intf is not rst and intf is not self)
            self.addSimpleParam(thisIf, "ASSOCIATED_BUSIF", ":".join(map(lambda intf: intf._name, intfs)))
            self.addSimpleParam(thisIf, "FREQ_HZ", str(DEFAULT_CLOCK))

class IP_Rst(IntfConfig):
    def __init__(self):
        super().__init__()
        self.name = "reset"
        self.version = "1.0"
        self.vendor = "xilinx.com" 
        self.library = "signal"
        self.map = "rst"
        
    def postProcess(self, component, entity, allInterfaces, thisIf):
        self.addSimpleParam(thisIf, "POLARITY", "ACTIVE_HIGH")

class IP_Rst_n(IntfConfig):
    def __init__(self):
        super().__init__()
        self.name = "reset"
        self.version = "1.0"
        self.vendor = "xilinx.com" 
        self.library = "signal"
        self.map = "rst"
        
    def postProcess(self, component, entity, allInterfaces, thisIf):
        self.addSimpleParam(thisIf, "POLARITY", "ACTIVE_LOW")

class IP_Handshake(IntfConfig):
    def __init__(self):
        super().__init__()
        self.name = "handshake"
        self.version = "1.0"
        self.vendor = "nic" 
        self.library = "user"
        self.map = {'rd': "ap_vld",
                     'rd': "ap_ack",
                     "data":"data"   }

class IP_BlockRamPort(IntfConfig):
    def __init__(self):
        super().__init__()
        self.name = "bram"
        self.version = "1.0"
        self.vendor = "xilinx.com"  
        self.library = "interface" 
        self.map = {'addr':"ADDR",
                    "clk": 'CLK',
                    'din':"DIN",
                    'dout':"DOUT",
                    'en': "EN",
                    'we': "WE",
                    } 



from hwt.hdlObjects.types.boolean import Boolean
from hwt.hdlObjects.types.bits import Bits
from hwt.simulator.types.simInt import SIM_INT
from hwt.hdlObjects.types.defs import INT
from hwt.hdlObjects.types.hdlType import HdlType
from hwt.bitmask import mask

def convertSimBits__val(self, sigOrVal, toType):
    if isinstance(toType, Boolean):
        return sigOrVal._eq(self.getValueCls().fromPy(1, self))
    elif isinstance(toType, Bits):
        if self.bit_length() == toType.bit_length():
            return sigOrVal._convSign(toType.signed)
    elif toType == INT or toType == SIM_INT:
        
        if self.signed:
            raise NotImplementedError()
        else:
            fullMask = mask(self.bit_length())
            return INT.getValueCls()(sigOrVal.val, INT, sigOrVal.vldMask == fullMask, sigOrVal.updateTime)
    return HdlType.defaultConvert(self, sigOrVal, toType)


from hwt.synthesizer.param import Param, getParam


class NotSpecified(Exception):
    """
    This error means that you need to implement this function to use this functionality
    
    f.e. you have to implement Simulation agent for interface when you create new one and you can not use existing
    """
    pass

def sameIntfAs(intf):
    _intf = intf.__class__()
    for p in intf._params:
        _intf._replaceParam(p._name, Param(getParam(p)))
    return _intf    

def walkPhysInterfaces(intf):
    if intf._interfaces:
        for si in intf._interfaces:
            yield from walkPhysInterfaces(si)
    else:
        yield intf


def forAllParams(intf, discovered=None):
    if discovered is None:
        discovered = set()
    
    for si in intf._interfaces:
        yield from forAllParams(si, discovered)
        
    for p in intf._params:
        if p not in discovered:
            discovered.add(p)
            yield p 



from hwt.pyUtils.nonRedefDict import NonRedefDict
from hwt.hdlObjects.reference import HdlRef
from hwt.hdlObjects.entity import Entity
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.hdlObjects.architecture import Architecture
from hwt.hdlObjects.function import Function
from hwt.hdlObjects.functionContainer import FunctionContainer

class RequireImportErr(Exception):
    """Is raised when program can not approach without importing specific reference"""
    def __init__(self, reference):
        super(RequireImportErr, self).__init__()
        self.reference = reference
        self.fileName = None
    
    def __repr__(self):
        return self.__str__()
    
    def __str__(self):
        if self.fileName:
            fileName = 'file %s' % self.fileName
        else:
            fileName = '' 
        return "<RequireImportErr %s require to import %s first>" % (fileName, str(self.reference))
    

class HdlContext(NonRedefDict):
    """
    Context of hdl, contains all hdl objects hidden behind references
    """
    def __init__(self, name, parent):
        self.name = name
        self.parent = parent
        
        self.entities = NonRedefDict()
        self.architectures = []
        self.packages = NonRedefDict()
        self.fileInfo = None
        
    def importLibFromGlobal(self, ref):
        """
        Import for example lib.package to local context
        """
        top = self
        while top.parent is not None:
            top = top.parent
        try:
            toImport = top
            for n in ref.names:
                toImport = toImport[n]
            if ref.all:
                for n in toImport:
                    self[n] = toImport[n]
            else:
                self[toImport.name] = toImport
        except KeyError:
            raise RequireImportErr(ref)
        
    def lookupGlobal(self, ref):
        """
        lookup reference upside down
        """
        p = self
        n = ref.names[0]  # [TODO]
        while p.parent is not None:
            p = p.parent
        if p is None:
            raise RequireImportErr(ref)
        try:
            for n in ref.names:
                p = p[n]
            return p
        except KeyError:
            raise RequireImportErr(ref)
        
    def lookupLocal(self, locRef):
        """[DEPRECATED]
        lookup only in this context
        """
        p = self
        n = locRef.names[-1]  # [TODO]
        while p is not None:
            try:
                return p[n]
            except KeyError:
                p = p.parent
        
        raise KeyError("Identificator %s not defined" % n)
    
    def insertObj(self, obj, caseSensitive, hierarchyOnly=False):
        """
        insert Entity, PackageHeader or Architecture in this context
        """
        from hwt.hdlObjects.package import PackageHeader  # [TODO] rm dependency
        def getName():
            n = obj.name
            if not caseSensitive:
                n = n.lower() 
            return n
        
        def insert(n):
            self.insert(HdlRef([n], caseSensitive), obj)
        
        if isinstance(obj, Entity):
            n = getName()
            self.entities[n] = obj
            insert(n)
        elif  isinstance(obj, RtlSignalBase):
            self[obj._name] = obj
        elif isinstance(obj, PackageHeader):
            n = getName()
            self.packages[n] = obj
            insert(n)
            
        elif isinstance(obj, Architecture):
            self.architectures.append(obj)
            
        elif isinstance(obj, Function):
            # functions are stored in FunctionContainer object
            n = getName()
            try:
                cont = self[n]
            except KeyError:
                cont = FunctionContainer(n, self)
                self.insert(HdlRef([n], caseSensitive), cont)
            cont.append(obj, suppressRedefinition=hierarchyOnly)
        else:
            raise NotImplementedError()
    
    def insert(self, ref, val):
        """
        insert any reference
        """
        c = self
        # for all names create or reuse hiearchy
        for n in ref.names[:-1]:
            c = c.setdefault(n, HdlContext(n, c))
        # store actual object at the end of hierarhy
        c[ref.names[-1]] = val
    
    def copyFrom(self, other):
        """
        copy everything from other HdlContext
        """
        for _, v in other.items():
            self.insertObj(v)
        self.parent = other.parent
        self.name = other.name
               
    def __str__(self):
        return "\n".join([
                    "\n".join([str(e) for _, e in self.entities.items()]),
                    "\n".join([str(a) for a in self.architectures]),
                    "\n".join([str(p) for p in self.packages]),
                    ])


from hwt.hdlObjects.architecture import Architecture
from hwt.hdlObjects.assignment import Assignment
from hwt.hdlObjects.entity import Entity
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.process import HWProcess
from hwt.hdlObjects.statements import IfContainer, WaitStm, \
    SwitchContainer
from hwt.hdlObjects.types.defs import BIT
from hwt.hdlObjects.value import Value
from hwt.synthesizer.assigRenderer import renderIfTree
from hwt.code import If
from hwt.synthesizer.exceptions import SigLvlConfErr
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.memory import RtlSyncSignal
from hwt.synthesizer.rtlLevel.optimalizator import removeUnconnectedSignals
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from hwt.synthesizer.rtlLevel.signalUtils.walkers import discoverSensitivity
from hwt.synthesizer.rtlLevel.utils import portItemfromSignal
from hwt.pyUtils.arrayQuery import where, distinctBy, groupedby


def isSignalHiddenInExpr(sig):
    """Some signals are just only connections in expression they done need to be rendered because
    they are hidden inside expression for example sig. from a+b in a+b+c"""
    if isinstance(sig, Value):
        return True
    try:
        d = sig.singleDriver()
        if isinstance(d, Operator):
            return True
    except MultipleDriversExc:
        pass
        
    return False

def _isEnclosed(objList):
    if not objList:
        return False
    for o in objList:
        if not isEnclosed(o):
            return False
    return True
    
def isEnclosed(obj):
    """
    Check if statement has any not used branch
    """
    if isinstance(obj, (Assignment, WaitStm)):
        return True
    elif isinstance(obj, IfContainer):
        for ol in [obj.ifTrue, obj.ifFalse]:
            if not _isEnclosed(ol):
                return False
        for _, ol in obj.elIfs:
            if not _isEnclosed(ol):
                return False
                
        return True 
    elif isinstance(obj, SwitchContainer):
        allCasesCovered = True
        for cond, ol in obj.cases:
            if cond is None:
                allCasesCovered = True
            if not _isEnclosed(ol):
                return False
            
        return allCasesCovered
    else:
        raise NotImplementedError(obj)


class RtlNetlist():
    """
    Container for signals and units
    @ivar signals: dict of all signals in context
    @ivar startsOfDataPaths:  is set of nodes where datapaths starts
    @ivar subUnits:           is set of all units in this context 
    """
    def __init__(self, parentForDebug=None):
        self.parentForDebug = parentForDebug
        self.globals = {}
        self.signals = set()
        self.startsOfDataPaths = set()
        self.subUnits = set()
        self.synthesised = False

    
    def sig(self, name, typ=BIT, clk=None, syncRst=None, defVal=None):
        """
        generate new signal in context
        @param clk: clk signal, if specified signal is synthesized as SyncSignal
        @param syncRst: reset 
        """
        if not isinstance(defVal, (Value, RtlSignal, InterfaceBase)):
            if isinstance(defVal, (InterfaceBase)):
                _defVal = defVal._sig
            else:    
                _defVal = typ.fromPy(defVal)
        else:
            _defVal = defVal._convert(typ)


        if clk is not None:
            s = RtlSyncSignal(self, name, typ, _defVal)
            if syncRst is not None and defVal is None:
                raise Exception("Probably forgotten default value on sync signal %s", name)
            if syncRst is not None:
                r = If(syncRst._isOn(),
                        [RtlSignal.__pow__(s, _defVal)]
                    ).Else(
                        [RtlSignal.__pow__(s, s.next)]
                    )
            else:
                r = [RtlSignal.__pow__(s, s.next)]
            
            If(clk._onRisingEdge(),
               r)
        else:
            if syncRst:
                raise SigLvlConfErr("Signal %s has reset but has no clk" % name)
            s = RtlSignal(self, name, typ, defaultVal=_defVal)
        
        self.signals.add(s)
        
        return s
   
    def buildProcessesOutOfAssignments(self):
        """
        Render conditional assignments to statements and wrap them with process statement
        """
        assigments = where(self.startsOfDataPaths,
                            lambda x: isinstance(x, Assignment)
                          )
        for sig, dps in groupedby(assigments, lambda x: x.dst):
            dps = list(dps)
            name = ""
            if not sig.hasGenericName:
                name = sig.name
            sig.hidden = False
            
            haveNotIndexes = True
            for dp in dps:
                haveNotIndexes = haveNotIndexes and not dp.indexes
            
            # render sequential statements in process
            # (conversion from netlist to statements)
            hasCombDriver = False
            for stm in renderIfTree(dps):
                
                p = HWProcess("assig_process_" + name)
                if sig._useNopVal and not isEnclosed(stm):
                    n = sig._nopVal
                    p.statements.append(Assignment(n, sig))
                    if isinstance(n, RtlSignal):
                        p.sensitivityList.add(n)
                    
                p.statements.append(stm)

                sensitivity = discoverSensitivity(stm)
                p.sensitivityList.update(sensitivity)

                isEventDependent = False
                for s in p.sensitivityList:
                    if isinstance(s, Operator):
                        # event operator
                        s.ops[0].hidden = False
                        isEventDependent = True
                    else:
                        s.hidden = False

                if hasCombDriver and not isEventDependent and haveNotIndexes:
                    raise MultipleDriversExc("Signal %s has multiple combinational drivers" % name)
                hasCombDriver = hasCombDriver or not isEventDependent
                
                yield p

    def mergeWith(self, other):
        """
        Merge two instances into this
        @attention: "others" becomes invalid because all signals etc. will be transferred into this 
        """
        assert not other.synthesised
        self.globals.update(other.globals)
        self.signals.update(other.signals)
        self.startsOfDataPaths.update(other.startsOfDataPaths)
        self.subUnits.update(other.subUnits)
        
        for s in other.signals:
            s.ctx = self
        

    def synthesize(self, name, interfaces):
        """
        Build Entity and architecture out of netlist representation
        """
        
        ent = Entity(name)
        ent._name = name + "_inst"  # instance name

        # create generics
        for _, v in self.globals.items():
            ent.generics.append(v)
        
        # create ports
        for s in interfaces:
            pi = portItemfromSignal(s, ent)
            pi.reigsterInternSig(s)
            ent.ports.append(pi)

        removeUnconnectedSignals(self)
        
        arch = Architecture(ent)
        for p in self.buildProcessesOutOfAssignments():
            arch.processes.append(p)
            

        # add signals, variables etc. in architecture
        for s in self.signals:
            if s not in interfaces and not s.hidden:
                arch.variables.append(s)
        
        # instanciate subUnits in architecture
        for u in self.subUnits:  
            arch.componentInstances.append(u) 
        
        # add components in architecture    
        for su in distinctBy(self.subUnits, lambda x: x.name):
            arch.components.append(su)
        
        self.synthesised = True
        
        return [ent, arch]
       



from hwt.hdlObjects.assignment import Assignment
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.statements import IfContainer
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.sliceVal import SliceVal
from hwt.hdlObjects.variables import SignalItem
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.vhdl.utils import VhdlVersion, vhdlTmplEnv
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc


IfTmpl = vhdlTmplEnv.get_template('if.vhd')
SwitchTmpl = vhdlTmplEnv.get_template('switch.vhd')

class DoesNotContainsTernary(Exception):
    pass

def ternaryOpsToIf(statements):
    """Convert all ternary operators to IfContainers"""
    stms = []
    
    for st in statements:
        if isinstance(st, Assignment):
            try:
                if not isinstance(st.src, RtlSignalBase):
                    raise DoesNotContainsTernary()
                d = st.src.singleDriver()
                if not isinstance(d, Operator) or d.operator != AllOps.TERNARY:
                    raise DoesNotContainsTernary()
                else:
                    ifc = IfContainer(d.ops[0],
                                      [Assignment(d.ops[1], st.dst)]
                                      ,
                                      [Assignment(d.ops[2], st.dst)]
                           )
                    stms.append(ifc)
                
            except (MultipleDriversExc, DoesNotContainsTernary):
                stms.append(st)
        else:
            stms.append(st)
    return stms


class VhdlSerializer_statements():
    
    @classmethod
    def WaitStm(cls, w, createTmpVarFn):
        if w.isTimeWait:
            return "wait for %d ns" % w.waitForWhat
        elif w.waitForWhat is None:
            return "wait"
        else:
            raise NotImplementedError()

    @classmethod
    def Assignment(cls, a, createTmpVarFn):
        dst = a.dst
        assert isinstance(dst, SignalItem)
        asHdl = lambda obj : cls.asHdl(obj, createTmpVarFn)
        valAsHdl = lambda v: cls.Value(v, createTmpVarFn)
        if dst.virtualOnly:
            symbol = ":="
        else:
            symbol = "<="
        
        if a.indexes is not None:
            for i in a.indexes:
                if isinstance(i, SliceVal):
                    i = i.clone()
                    i.val = (i.val[0] + 1, i.val[1])
                dst = dst[i]   
        
                    
        if dst._dtype == a.src._dtype:
            return "%s %s %s" % (asHdl(dst), symbol, valAsHdl(a.src))
        else:
            srcT = a.src._dtype
            dstT = dst._dtype
            if isinstance(srcT, Bits) and isinstance(dstT, Bits):
                sLen = srcT.bit_length()
                dLen = dstT.bit_length()
                if sLen == dLen:
                    if sLen == 1 and srcT.forceVector != dstT.forceVector:
                        if srcT.forceVector:
                            return "%s %s %s(0)" % (asHdl(dst), symbol, valAsHdl(a.src)) 
                        else:
                            return "%s(0) %s %s" % (asHdl(dst), symbol, valAsHdl(a.src)) 
                    elif srcT.signed is not dstT.signed:
                        return "%s %s %s" % (asHdl(dst), symbol, valAsHdl(a.src._convSign(dstT.signed)))
            
            raise SerializerException("%s %s %s  is not valid assignment\n because types are different (%s; %s) " % 
                         (asHdl(dst), symbol, valAsHdl(a.src), repr(dst._dtype), repr(a.src._dtype)))

    @classmethod
    def IfContainer(cls, ifc, createTmpVarFn):
        asHdl = lambda obj : cls.asHdl(obj, createTmpVarFn)
        cond = cls.condAsHdl(ifc.cond, True, createTmpVarFn)
        elIfs = []
        if cls.VHDL_VER < VhdlVersion.v2008:
            ifTrue = ternaryOpsToIf(ifc.ifTrue)
            ifFalse = ternaryOpsToIf(ifc.ifFalse)
        else:
            ifTrue = ifc.ifTrue
            ifFalse = ifc.ifFalse
        
        for c, statements in ifc.elIfs:
            if cls.VHDL_VER < VhdlVersion.v2008:
                statements = ternaryOpsToIf(statements)
                
            elIfs.append((cls.condAsHdl(c, True, createTmpVarFn), [asHdl(s) for s in statements]))
        
        return IfTmpl.render(cond=cond,
                             ifTrue=[asHdl(s) for s in ifTrue],
                             elIfs=elIfs,
                             ifFalse=[asHdl(s) for s in ifFalse])  
    
    @classmethod
    def SwitchContainer(cls, sw, createTmpVarFn):
        asHdl = lambda obj : cls.asHdl(obj, createTmpVarFn)
        switchOn = cls.condAsHdl(sw.switchOn, False, createTmpVarFn)
        
        cases = []
        for key, statements in sw.cases:
            if key is not None:  # None is default
                key = cls.asHdl(key, createTmpVarFn)
                
            if cls.VHDL_VER < VhdlVersion.v2008:
                statements = ternaryOpsToIf(statements)
                
            cases.append((key, [asHdl(s) for s in statements]))  
        return SwitchTmpl.render(switchOn=switchOn,
                                 cases=cases)  
    
    @classmethod
    def MapExpr(cls, m, createTmpVar):
        return   "%s => %s" % (m.compSig.name, cls.asHdl(m.value, createTmpVar))



from hwt.hdlObjects.variables import SignalItem

class SimSignal(SignalItem):
    """
    Class of signal simulation functions
    
    @ivar _writeCallbacks: list of callback functions(signal, simulator) which is called
                           when new (changed) value is written to this signal
    """
    __slots__ = ["name", "_val", "_oldVal", "_writeCallbacks",
                 "simSensProcs", "simRisingSensProcs", "simFallingSensProcs"]
    def __init__(self, ctx, name, dtype, defaultVal=None):
        ctx.signals.add(self)
        self.hidden = False
        self._writeCallbacks = []
        self.simSensProcs = set()
        self.simRisingSensProcs = set()
        self.simFallingSensProcs = set()
        super(SimSignal, self).__init__(name, dtype, defaultVal)
        
    
    def simPropagateChanges(self, simulator):
        v = self._val
        self._oldVal = v

        # run all sensitive processes    
        log = simulator.config.logPropagation
        for p in self.simSensProcs:        
            if log:
                log(simulator, self, p)
                
            simulator.addHwProcToRun(self, p)

        if self.simRisingSensProcs or self.simFallingSensProcs: 
            if v.val or not v.vldMask:
                for p in self.simRisingSensProcs:        
                    if log:
                        log(simulator, self, p)
                
                    simulator.addHwProcToRun(self, p)
                    
            if not v.val or not v.vldMask:
                for p in self.simFallingSensProcs:        
                    if log:
                        log(simulator, self, p)
                
                    simulator.addHwProcToRun(self, p)
    
    def simUpdateVal(self, simulator, valUpdater):
        """
        Method called by simulator to update new value for this object
        """
        
        dirtyFlag, newVal = valUpdater(self._oldVal)
        self._val = newVal
        newVal.updateTime = simulator.now
        
        if dirtyFlag:
            log = simulator.config.logChange
            if  log:
                log(simulator.now, self, newVal)
            
            # run write callbacks we have to create new list to allow 
            # registering of new call backs in callbacks
            callBacks = self._writeCallbacks
            self._writeCallbacks = []
            for c in callBacks:
                # simulation processes
                simulator.process(c(simulator))

            self.simPropagateChanges(simulator)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class Entity(object):
    """
    Hdl container of hdl configuration and interfaces
    """
    def __init__(self, name):
        self.name = name
        self.origin = None  # creator of this object
        self.generics = []
        self.ports = []
        self.ctx = {}
        self.discovered = False

    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer
        return VhdlSerializer.Entity(self, VhdlSerializer.getBaseNameScope())
    




class HWProcess():
    """
    Hdl process container
    """
    def __init__(self, name):
        self.name = name
        self.statements = []
        self.sensitivityList = set()

    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer
        return VhdlSerializer.formater(VhdlSerializer.HWProcess(self, VhdlSerializer.getBaseNameScope()))



from hwt.simulator.agentBase import SyncAgentBase


class FifoReaderAgent(SyncAgentBase):
    
    def __init__(self, intf, clk=None, rstn=None, allowNoReset=False):
        super(FifoReaderAgent, self).__init__(intf, clk, rstn, allowNoReset)
        self.data = []
        self.readPending = False
        
    def monitor(self, s):
        intf = self.intf
        if self.readPending:
            yield s.updateComplete
            d = s.read(intf.data)
            self.data.append(d)
            self.readPending = False
        if s.r(self.rst_n).val and self.enable:
            
            rd = not s.r(intf.wait).val
            s.w(rd, intf.en)
            
            if rd:
                self.readPending = True
            #    self.readPending
            #    yield s.updateComplete
            #    d = s.read(intf.data)
            #    self.data.append(d)
        else:
            s.w(0, intf.en)
            
    def driver(self, s):
        raise NotImplementedError()

class FifoWriterAgent(SyncAgentBase):
        
    def monitor(self, s):
        raise NotImplementedError()
            
    def driver(self, s):
        intf = self.intf
        
        if s.r(self.rst_n).val and not s.r(intf.wait).val \
           and self.data and self.enable:
            s.w(self.data.pop(0), intf.data)
            s.w(1, intf.en)
        else:
            s.w(None, intf.data)
            s.w(0, intf.en)



from hwt.hdlObjects.types.integer import Integer


class SimInteger(Integer):
    @classmethod
    def getConvertor(cls):
        from hwt.simulator.types.simIntConversions import convertSimInteger__val
        return convertSimInteger__val        


SIM_INT = SimInteger()
# create hdl integer value (for example integer value in vhdl)
simHInt = lambda val: SIM_INT.fromPy(val)


from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.defs import BOOL, INT
from hwt.hdlObjects.types.slice import Slice
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.hdlObjects.value import Value
from hwt.synthesizer.param import evalParam
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase

class ArrayVal(Value):
    """
    Class of value of array
    """
    
    @classmethod
    def fromPy(cls, val, typeObj):
        size = evalParam(typeObj.size)
        if isinstance(size, Value):
            size = size.val
        
        if val is None:
            v = typeObj.elmType.fromPy(None)
            elements = [v.clone() for _ in range(size)]
        else:
            elements = []
            for v in val:
                if isinstance(v, RtlSignalBase):  # is signal
                    assert v._dtype == typeObj.elmType
                    e = v
                else:
                    e = typeObj.elmType.fromPy(v)
                elements.append(e)
        
        
        return cls(elements, typeObj, int(val is not None))
    
    def _getitem__val(self, key):
        v = self.val[key.val].clone()
        if not key._isFullVld():
            v.vldMask = 0
        
        return v
    
    def __getitem__(self, key):
        iamVal = isinstance(self, Value)
        key = toHVal(key)
        isSLICE = isinstance(key, Slice.getValueCls())
        
        if isSLICE:
            raise NotImplementedError()
        elif isinstance(key, RtlSignalBase):
            key = key._convert(INT)
        elif isinstance(key, Value):
            pass
        else:
            raise NotImplementedError("Index operation not implemented for index %s" % 
                                       (repr(key)))
            
        if iamVal and isinstance(key, Value):
            return self._getitem__val(key)
        
        return Operator.withRes(AllOps.INDEX, [self, key], self._dtype.elmType)
    
    def _setitem__val(self, index, value):
        self.updateTime = max(index.updateTime, value.updateTime)
        if index._isFullVld():
            self.val[index.val] = value.clone()
        else:
            for v in self.val:
                v.vldMask = 0 
                v.updateTime = self.updateTime
            self.vldMask = 0
    
    
    def __setitem__(self, index, value):
        assert isinstance(self, Value)
        assert index._dtype == INT, index._dtype 
        return self._setitem__val(index, value)
        
    def _eq__val(self, other):
        assert self._dtype.elmType == other._dtype.elmType
        assert self._dtype.size == other._dtype.size
        
        eq = True
        first = self.val[0]
        vld = first.vldMask
        updateTime = first.updateTime
        
        for a, b in zip(self.val, other.val):
            eq = eq and a == b
            vld = vld & a.vldMask & b.vldMask
            updateTime = max(updateTime, a.updateTime, b.updateTime)
        return BOOL.getValueCls()(eq, BOOL, vld, updateTime)
        
    def _eq(self, other):
        assert isinstance(other, ArrayVal)
        return self._eq__val(other)





def removeUnconnectedSignals(netlist):
    
    toDelete = set()
    while True:
        for sig in netlist.signals:
            if not sig.endpoints:
                try:
                    if sig._interface is not None:
                        continue
                except AttributeError:
                    pass
                toDelete.add(sig)
                
        if toDelete:        
            for sig in toDelete:
                netlist.signals.remove(sig)
            toDelete = set()
        else:
            return


from collections import namedtuple

from hwt.hdlObjects.architecture import Architecture
from hwt.hdlObjects.assignment import Assignment
from hwt.hdlObjects.entity import Entity
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.statements import IfContainer, \
    SwitchContainer, WhileContainer, WaitStm
from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.defs import BOOL, BIT
from hwt.hdlObjects.types.enum import Enum
from hwt.hdlObjects.types.hdlType import InvalidVHDLTypeExc
from hwt.hdlObjects.value import Value
from hwt.hdlObjects.variables import SignalItem
from hwt.pyUtils.arrayQuery import arr_any, groupedby
from hwt.serializer.constants import SERI_MODE
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.nameScope import LangueKeyword, NameScope
from hwt.serializer.serializerClases.mapExpr import MapExpr
from hwt.serializer.serializerClases.portMap import PortMap 
from hwt.serializer.utils import maxStmId
from hwt.serializer.vhdl.formater import formatVhdl
from hwt.serializer.vhdl.ops import VhdlSerializer_ops
from hwt.serializer.vhdl.statements import VhdlSerializer_statements
from hwt.serializer.vhdl.types import VhdlSerializer_types
from hwt.serializer.vhdl.utils import VhdlVersion, vhdlTmplEnv
from hwt.serializer.vhdl.value import VhdlSerializer_Value
from hwt.synthesizer.interfaceLevel.unit import Unit
from hwt.synthesizer.param import getParam, evalParam
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


architectureTmpl = vhdlTmplEnv.get_template('architecture.vhd')
entityTmpl = vhdlTmplEnv.get_template('entity.vhd')
processTmpl = vhdlTmplEnv.get_template('process.vhd')
componentTmpl = vhdlTmplEnv.get_template('component.vhd')
componentInstanceTmpl = vhdlTmplEnv.get_template('component_instance.vhd')

VHLD_KEYWORDS = [
    "abs", "access", "across", "after", "alias", "all", "and", "architecture", "array",
    "assert", "attribute", "begin", "block", "body", "break", "bugger", "bus", "case",
    "component", "configuration", "constant", "disconnect", "downto", "end", "entity",
    "else", "elsif", "exit", "file", "for", "function", "generate", "generic", "group",
    "guarded", "if", "impure", "in", "inertial", "inout", "is", "label", "library", "limit",
    "linkage", "literal", "loop", "map", "mod", "nand", "nature", "new", "next", "noise",
    "nor", "not", "null", "of", "on", "open", "or", "others", "out", "package", "port",
    "postponed", "process", "procedure", "procedural", "pure", "quantity", "range",
    "reverse_range", "reject", "rem", "record", "reference", "register", "report", "return",
    "rol", "ror", "select", "severity", "shared", "signal", "sla", "sll", "spectrum", "sra",
    "srl", "subnature", "subtype", "terminal", "then", "through", "to", "tolerance", "transport",
    "type", "unaffected", "units", "until", "use", "variable", "wait", "with", "when", "while",
    "xnor", "xor"]        


def freeze_dict(data):
    keys = sorted(data.keys())
    frozen_type = namedtuple(''.join(keys), keys)
    return frozen_type(**data)

def paramsToValTuple(unit):
    d = {}
    for p in unit._params:
        name = p.getName(unit)
        v = evalParam(p)
        d[name] = v
    return freeze_dict(d)

def onlyPrintDefaultValues(suggestedName, dtype):
    # [TODO] it is better to use RtlSignal
    s = SignalItem(suggestedName, dtype, virtualOnly=True)
    s.hidden = False
    serializedS = VhdlSerializer.SignalItem(s, onlyPrintDefaultValues, declaration=True)
    print(serializedS)
    return s

class VhdlSerializer(VhdlSerializer_Value, VhdlSerializer_ops, VhdlSerializer_types, VhdlSerializer_statements):
    VHDL_VER = VhdlVersion.v2002
    __keywords_dict = {kw: LangueKeyword() for kw in VHLD_KEYWORDS}
    formater = formatVhdl
    fileExtension = '.vhd'
    
    @classmethod
    def getBaseNameScope(cls):
        s = NameScope(True)
        s.setLevel(1)
        s[0].update(cls.__keywords_dict)
        return s
    
    @classmethod
    def serializationDecision(cls, obj, serializedClasses, serializedConfiguredUnits):
        """
        Decide if this unit should be serialized or not eventually fix name to fit same already serialized unit
        @param serializedClasses: unitCls : unitobj
        @param serializedConfiguredUnits: (unitCls, paramsValues) : unitObj
                                          where paramsValues are named tuple name:value
        """
        isEnt = isinstance(obj, Entity) 
        isArch = isinstance(obj, Architecture)
        if isEnt:
            unit = obj.origin
        elif isArch:
            unit = obj.entity.origin
        else:
            return True
        
        assert isinstance(unit, Unit)
        m = unit._serializerMode
        
        if m == SERI_MODE.ALWAYS:
            return True
        elif m == SERI_MODE.ONCE:
            if isEnt:
                try:
                    prevUnit = serializedClasses[unit.__class__]
                except KeyError:
                    serializedClasses[unit.__class__] = unit
                    obj.name = unit.__class__.__name__
                    return True

                obj.name = prevUnit._entity.name
                return False
            
            return serializedClasses[unit.__class__] is unit
        elif m == SERI_MODE.PARAMS_UNIQ:
            params = paramsToValTuple(unit)
            k = (unit.__class__, params)
            if isEnt:
                try:
                    prevUnit = serializedConfiguredUnits[k]
                except KeyError:
                    serializedConfiguredUnits[k] = unit
                    return True

                obj.name = prevUnit._entity.name
                return False
            
            return serializedConfiguredUnits[k] is unit    
        elif m == SERI_MODE.EXCLUDE:
            if isEnt:
                obj.name = unit.__class__.__name__
            return False
        else:
            raise NotImplementedError("Not implemented serializer mode %r on unit %r" % (m, unit))
    
    @classmethod
    def asHdl(cls, obj, createTmpVarFn):
        """
        Convert object to VHDL string
        @param obj: object to serialize
        @param createTmpVarFn: function (sugestedName, dtype) returns variable
                            this function will be called to create tmp variables
        """
        if hasattr(obj, "asVhdl"):
            return obj.asVhdl(cls, createTmpVarFn)
        elif isinstance(obj, RtlSignalBase):
            return cls.SignalItem(obj, createTmpVarFn)
        elif isinstance(obj, Value):
            return cls.Value(obj, createTmpVarFn)
        else:
            try:
                serFn = getattr(cls, obj.__class__.__name__)
            except AttributeError:
                raise NotImplementedError("Not implemented for %s" % (repr(obj)))
            return serFn(obj, createTmpVarFn)
    
    @classmethod
    def FunctionContainer(cls, fn):
        return fn.name
    
    @classmethod
    def Architecture(cls, arch, scope):
        variables = []
        procs = []
        extraTypes = set()
        extraTypes_serialized = []
        arch.variables.sort(key=lambda x: x.name)
        arch.processes.sort(key=lambda x: (x.name, maxStmId(x)))
        arch.components.sort(key=lambda x: x.name)
        arch.componentInstances.sort(key=lambda x: x._name)
        def createTmpVarFn(suggestedName, dtype):
            raise NotImplementedError()
        
        for v in arch.variables:
            t = v._dtype
            # if type requires extra definition
            if isinstance(t, (Enum, Array)) and t not in extraTypes:
                extraTypes.add(v._dtype)
                extraTypes_serialized.append(cls.HdlType(t, createTmpVarFn, scope, declaration=True))

            v.name = scope.checkedName(v.name, v)
            serializedVar = cls.SignalItem(v, createTmpVarFn, declaration=True)
            variables.append(serializedVar)
            
        
        for p in arch.processes:
            procs.append(cls.HWProcess(p, scope))
        
        # architecture names can be same for different entities
        # arch.name = scope.checkedName(arch.name, arch, isGlobal=True)    
             
             
        uniqComponents = map(lambda x: x[1][0], groupedby(arch.components, lambda c: c.name))
        
        return architectureTmpl.render({
        "entityName"         :arch.getEntityName(),
        "name"               :arch.name,
        "variables"          :variables,
        "extraTypes"         :extraTypes_serialized,
        "processes"          :procs,
        "components"         :map(lambda c: cls.Component(c, createTmpVarFn),
                                   uniqComponents),
        "componentInstances" :map(lambda c: cls.ComponentInstance(c, createTmpVarFn, scope),
                                   arch.componentInstances)
        })
        
    @classmethod
    def comment(cls, comentStr):
        return "--" + comentStr.replace("\n", "\n--")
    
    @classmethod
    def Component(cls, entity, createTmpVar):
        entity.ports.sort(key=lambda x: x.name)
        entity.generics.sort(key=lambda x: x.name)
        return componentTmpl.render({
                "ports": [cls.PortItem(pi, createTmpVar) for pi in entity.ports],
                "generics": [cls.GenericItem(g, createTmpVar) for g in entity.generics],
                "entity": entity
                })      

    @classmethod
    def ComponentInstance(cls, entity, createTmpVarFn, scope):
        portMaps = []
        for pi in entity.ports:
            pm = PortMap.fromPortItem(pi)
            portMaps.append(pm)
        
        genericMaps = []
        for g in entity.generics:
            gm = MapExpr(g, g._val)
            genericMaps.append(gm) 
        
        if len(portMaps) == 0:
            raise Exception("Incomplete component instance")
        
        # [TODO] check component instance name
        return componentInstanceTmpl.render({
                "instanceName" : entity._name,
                "entity": entity,
                "portMaps": [cls.PortConnection(x, createTmpVarFn) for x in portMaps],
                "genericMaps" : [cls.MapExpr(x, createTmpVarFn) for x in genericMaps]
                })     

    @classmethod
    def Entity(cls, ent, scope):
        ports = []
        generics = []
        ent.ports.sort(key=lambda x: x.name)
        ent.generics.sort(key=lambda x: x.name)
        
        def createTmpVarFn(suggestedName, dtype):
            raise NotImplementedError()

        ent.name = scope.checkedName(ent.name, ent, isGlobal=True)
        for p in ent.ports:
            p.name = scope.checkedName(p.name, p)
            ports.append(cls.PortItem(p, createTmpVarFn))
            
        for g in ent.generics:
            g.name = scope.checkedName(g.name, g)
            generics.append(cls.GenericItem(g, createTmpVarFn))    

        entVhdl = entityTmpl.render({
                "name": ent.name,
                "ports" : ports,
                "generics" : generics
                })

        doc = ent.__doc__
        if doc:
            doc = cls.comment(doc) + "\n"
            return doc + entVhdl   
        else:
            return entVhdl
    
    @classmethod
    def condAsHdl(cls, cond, forceBool, createTmpVarFn):
        if isinstance(cond, RtlSignalBase):
            cond = [cond]
        else:
            cond = list(cond)
        if len(cond) == 1:
            c = cond[0]
            if not forceBool or c._dtype == BOOL:
                return cls.asHdl(c, createTmpVarFn)
            elif c._dtype == BIT:
                return "(" + cls.asHdl(c, createTmpVarFn) + ")=" + cls.BitLiteral(1, 1) 
            elif isinstance(c._dtype, Bits):
                width = c._dtype.bit_length()
                return "(" + cls.asHdl(c, createTmpVarFn) + ")/=" + cls.BitString(0, width)
            else:
                raise NotImplementedError()
            
        else:
            return " AND ".join(map(lambda x: cls.condAsHdl(x, forceBool, createTmpVarFn), cond))
    
      
    @classmethod
    def GenericItem(cls, g, createTmpVarFn):
        s = "%s : %s" % (g.name, cls.HdlType(g._dtype, createTmpVarFn))
        if g.defaultVal is None:
            return s
        else:  
            return  "%s := %s" % (s, cls.Value(getParam(g.defaultVal).staticEval(), createTmpVarFn))
    
    @classmethod
    def PortConnection(cls, pc, createTmpVarFn):
        if pc.portItem._dtype != pc.sig._dtype:
            raise SerializerException("Port map %s is nod valid (types does not match)  (%s, %s)" % (
                      "%s => %s" % (pc.portItem.name, cls.asHdl(pc.sig, createTmpVarFn)),
                      repr(pc.portItem._dtype), repr(pc.sig._dtype)))
        return " %s => %s" % (pc.portItem.name, cls.asHdl(pc.sig, createTmpVarFn))      
    
    @classmethod
    def DIRECTION(cls, d):
        return d.name

    
    @classmethod
    def PortItem(cls, pi, createTmpVarFn):
        try:
            return "%s : %s %s" % (pi.name, cls.DIRECTION(pi.direction),
                                   cls.HdlType(pi._dtype, createTmpVarFn))
        except InvalidVHDLTypeExc as e:
            e.variable = pi
            raise e
    
    @classmethod
    def sensitivityListItem(cls, item, createTmpVarFn):
        if isinstance(item, Operator):
            item = item.ops[0]
        
        return cls.asHdl(item, createTmpVarFn)
            

    @classmethod
    def HWProcess(cls, proc, scope):
        """
        Serialize HWProcess objects as VHDL
        @param scope: name scope to prevent name collisions
        """
        body = proc.statements
        extraVars = []
        extraVarsSerialized = []

        def createTmpVarFn(suggestedName, dtype):
            # [TODO] it is better to use RtlSignal
            s = SignalItem(None, dtype, virtualOnly=True)
            s.name = scope.checkedName(suggestedName, s)
            s.hidden = False
            serializedS = cls.SignalItem(s, createTmpVarFn, declaration=True)
            extraVars.append(s)
            extraVarsSerialized.append(serializedS)
            return s
        

        sensitivityList = sorted(map(lambda s: cls.sensitivityListItem(s, None), proc.sensitivityList))
        statemets = [ cls.asHdl(s, createTmpVarFn) for s in body]
        
        hasToBeVhdlProcess = extraVars or arr_any(body, lambda x: isinstance(x,
                                        (IfContainer, SwitchContainer, WhileContainer, WaitStm)))
        
        if hasToBeVhdlProcess:
            proc.name = scope.checkedName(proc.name, proc)
        
        extraVarsInit = []
        for s in extraVars:
            a = Assignment(s.defaultVal, s, virtualOnly=True)
            extraVarsInit.append(cls.Assignment(a, createTmpVarFn))

        return processTmpl.render({
              "name": proc.name,
              "hasToBeVhdlProcess": hasToBeVhdlProcess,
              "extraVars": extraVarsSerialized,
              "sensitivityList": ", ".join(sensitivityList),
              "statements": extraVarsInit + statemets })



from hwt.hdlObjects.types.defs import SLICE
from hwt.simulator.simSignal import SimSignal
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class IndexSimSignalProxy(SimSignal):
    """
    Proxy which allows place indexing operations on signal
    """
    def __init__(self, name, baseSignal, dtype, upperIndex, lowerIndex=None):
        """
        @param lowerIndex: if this is none only upper index will be used like sig[upperIndex]
                            else range select like sig[upperIndex:lowerIndex]
        """
        defaultVal = dtype.fromPy(None) 

        self.name = name
        self._dtype = dtype
        self.defaultVal = defaultVal

        self.hidden = False
        self._writeCallbacks = []
        self.simSensProcs = set()
        self.simRisingSensProcs = set()
        self.simFallingSensProcs = set()
        self._signal = baseSignal
        if lowerIndex is None:
            self.__index = toHVal(upperIndex)
        else:
            self.__index = SLICE.fromPy([upperIndex - 1, lowerIndex])
        self._setDefValue()
        
    def _generic_val_get(self, v):
        return v._getitem__val(self.__index)
    
    def _val_get(self):
        return self._generic_val_get(self._signal._val)
    
    def _oldVal_get(self):
        return self._generic_val_get(self._signal._oldVal)
    
    _val = property(_val_get)
    _oldVal = property(_oldVal_get)

    def simPropagateChanges(self, simulator):
        raise NotImplementedError("Call this function on real signals not on this proxy")
    
    def simUpdateVal(self, simulator, valUpdater):
        """
        Method called by simulator to update new value for this object
        we are only delegating update on parent signal
        """
        dirtyFlag, newVal = valUpdater(self._oldVal)
        newVal.updateTime = simulator.now
        
        if dirtyFlag:
            # run write callbacks we have to create new list to allow 
            # registering of new call backs in callbacks
            callBacks = self._writeCallbacks
            self._writeCallbacks = []
            for c in callBacks:
                # simulation processes
                simulator.process(c(simulator))
                
            def updateParent(oldVal):
                oldVal[self.__index] = newVal
                return True, oldVal
            
            self._signal.simUpdateVal(simulator, updateParent)
    
    def _setDefValue(self):
        v = self.defaultVal
        if isinstance(v, RtlSignalBase):
            v = v.staticEval()
        
        s = self._signal
        s._val[self.__index] = v.clone()
        s._oldVal[self.__index] = self._val.clone()
        s._oldVal.vldMask = 0



from hwt.hdlObjects.typeShortcuts import vecT
from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.arrayVal import ArrayVal
from hwt.hdlObjects.types.bitsConversions import convertBits__val
from hwt.hdlObjects.types.bitsVal import BitsVal
from hwt.hdlObjects.types.defs import BIT, SLICE
from hwt.hdlObjects.types.enum import Enum
from hwt.hdlObjects.constants import DIRECTION, SENSITIVITY
from hwt.simulator.types.simIntConversions import convertSimInteger__val 
from hwt.simulator.types.simInt import SIM_INT, simHInt
from hwt.simulator.types.simBits import simBitsT
from hwt.simulator.types.simBitsConversions import convertSimBits__val
from hwt.simulator.simModel import (SimModel, sensitivity, connectSimPort,
                                            simEvalCond, mkUpdater, mkArrayUpdater)
from hwt.code import Concat, power
from hwt.synthesizer.rtlLevel.netlist import RtlNetlist
from hwt.simulator.simSignal import SimSignal
from hwt.hdlObjects.types.sliceVal import SliceVal

{% for c in componentInstances %}
if "{{c.name}}" not in locals(): # support for all models in single file
    from {{c.name}} import {{c.name}}{% endfor %}
{% for imp in imports %}
{{imp}}
{% endfor %}

class {{ name }}(SimModel):
    _name = "{{ name }}" 
    _cntx = RtlNetlist(){% for t in extraTypes %} 
    {{t}}{% endfor %}
    
    # ports{% for name, dtype in ports %}
    {{name}} = SimSignal(_cntx, "{{name}}", {{dtype}}){% endfor %}
    
    # internal signals{% for name, dtype, defVal in signals %}
    {{name}} = SimSignal(_cntx, "{{name}}", {{dtype}}, defaultVal={{defVal}}){% endfor %}
    
{% for proc in processes %}
{{proc}}
{% endfor %}
    
    def __init__(self):
        self._interfaces = [{% for name, _ in ports   %}self.{{name}},
                            {% endfor %}{% for name, _, _ in signals %}self.{{name}},
                            {% endfor %}]
        self._processes = [{% for procName in processesNames %}self.{{procName}},
                           {% endfor %}]
        {% for c in componentInstances %}
        # connect ports{% for p in c.ports %}
        connectSimPort(self, {{c.name}},"{{p.src.name}}", "{{p.dst.name}}", {{p.direction}}){% endfor %}
        self.{{c._name}} = {{c.name}}()
        {% endfor %}
        
        self._units = [{% for c in componentInstances %}self.{{c._name}},
                       {% endfor %}
                       ]
        {% for proc in processObjects %}
        sensitivity(self.{{proc.name}}, {% 
            for s in proc.sensitivityList %}{% 
                if isOp(s) %}({{ sensitivityByOp(s.operator) }}, self.{{s.ops[0].name}}){% 
                else %}self.{{s.name}}{%
                endif %}{% 
                if not loop.last %}, {% 
                endif %}{% 
            endfor %}){% 
        endfor %}




class IfTreeNode():
    """
    
    @ival neg: dict condition : IfTreeNode object
    @ivar negSt: statements which will happen if condition is not met
    
    """
    def __init__(self, cond):
        self.cond = cond
        self.pos = []
        self.neg = []






from jinja2.loaders import PackageLoader
from jinja2.environment import Environment

vhdlTmplEnv = Environment(loader=PackageLoader('hwt', 'serializer/vhdl/templates'))

class VhdlVersion():
    v2002 = 2002
    v2008 = 2008


from hwt.synthesizer.interfaceLevel.interface import Interface
from hwt.synthesizer.param import Param
from hwt.interfaces.std import Signal
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.constants import DIRECTION

class TristateSig(Interface):
    """
    Tristate interface
    in order to make this a vector[0] instead of single bit use forceVector=True
    """
    def _config(self):
        self.DATA_WIDTH = Param(1)
        self.forceVector = False
        
    def _declr(self):
        t = Bits((self.DATA_WIDTH-1)._downto(0), self.forceVector)
        
        self.t = Signal(dtype=t) # connect
        self.i = Signal(dtype=t, masterDir=DIRECTION.IN) # input
        self.o = Signal(dtype=t) # output

class TristateClk(TristateSig):
    pass


from hwt.hdlObjects.assignment import Assignment
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.defs import BOOL
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.hdlObjects.value import Value
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc


def tv(signal):
    """
    Value class for type of signal
    """
    return signal._dtype.getValueCls()

class RtlSignalOps():
    def _convert(self, toT):
        return tv(self)._convert(self, toT)
    
    def naryOp(self, operator, opCreateDelegate, *otherOps):
        k = (operator, *otherOps)
        try:
            return self._usedOps[k]
        except KeyError:
            o = opCreateDelegate(self, *otherOps)
            self._usedOps[k] = o
            return o
        
        return o
    
    def __invert__(self):
        return self.naryOp(AllOps.NOT, tv(self).__invert__)
        
    def _onRisingEdge(self, now=None):
        return self.naryOp(AllOps.RISING_EDGE, tv(self)._onRisingEdge, now)
    
    def _onFallingEdge(self, now=None):
        return self.naryOp(AllOps.FALLIGN_EDGE, tv(self)._onFallingEdge, now)
    
    def _hasEvent(self, now=None):
        raise self.naryOp(AllOps.EVENT, tv(self)._hasEvent, now)
    
    def _isOn(self):
        return self._dtype.convert(self, BOOL)
    
    
    # conversions
    def _convSign(self, signed):
        return tv(self)._convSign(self, signed)
    
    def _signed(self):
        return tv(self)._signed(self)
    
    def _unsigned(self):
        return tv(self)._unsigned(self)
    
    def _vec(self):
        return tv(self)._vec(self)
       
    # logic    
    def __and__(self, other):
        return self.naryOp(AllOps.AND_LOG, tv(self).__and__, other)
    
    def __xor__(self, other):
        return self.naryOp(AllOps.XOR, tv(self).__xor__, other)

    def __or__(self, other):
        return self.naryOp(AllOps.OR_LOG, tv(self).__or__, other)

    # cmp
    def _eq(self, other):
        """__eq__ is not overloaded because it will destroy hashability of object"""
        return self.naryOp(AllOps.EQ, tv(self)._eq, other)

    def __ne__(self, other):
        return self.naryOp(AllOps.NEQ, tv(self).__ne__, other)
    
    def __ge__(self, other):
        return self.naryOp(AllOps.GE, tv(self).__ge__, other)
    
    def __gt__(self, other):
        return self.naryOp(AllOps.GREATERTHAN, tv(self).__gt__, other)
    
    def __lt__(self, other):
        return self.naryOp(AllOps.LOWERTHAN, tv(self).__lt__, other)
    
    def __le__(self, other):
        return self.naryOp(AllOps.LE, tv(self).__le__, other)
    
    
    # arithmetic
    def __add__(self, other):
        return self.naryOp(AllOps.ADD, tv(self).__add__, other)
    
    def __sub__(self, other):
        return self.naryOp(AllOps.SUB, tv(self).__sub__, other)
    
    def __mul__(self, other):
        return self.naryOp(AllOps.MUL, tv(self).__mul__, other)

    def __floordiv__(self, divider):
        return self.naryOp(AllOps.DIV, tv(self).__floordiv__, divider)
    
    # selections
    def _downto(self, to):
        return self.naryOp(AllOps.DOWNTO, tv(self)._downto, to)
    
    def __getitem__(self, key):
        operator = AllOps.INDEX
        if isinstance(key, slice):
            hashableKey = (key.start, key.stop, key.step)
        else:
            hashableKey = key
        k = (operator, hashableKey)
        try:
            return self._usedOps[k]
        except KeyError:
            o = tv(self).__getitem__(self, key)
            self._usedOps[k] = o
            return o  
        
        return o

    def _reversed(self):
        length = self._dtype.bit_length()
        concated = None
        for i in range(length):
            if i == 0:
                concated = self[i]
            else:
                concated = concated._concat(self[i])
        return concated

    def _concat(self, *operands):
        return self.naryOp(AllOps.CONCAT, tv(self)._concat, *operands)
    
    def _ternary(self, ifTrue, ifFalse):
        return self.naryOp(AllOps.TERNARY, tv(self)._ternary, ifTrue, ifFalse)
    
    def _getIndexCascade(self):
        """
        Find out if this signal is something indexed
        """
        try:
            # now I am result of the index  xxx[xx] <= source
            # get index op
            d = self.singleDriver()
            try:
                op = d.operator
            except AttributeError:
                op = None
            if op == AllOps.INDEX:
                # get signal on which is index applied
                indexedOn = d.ops[0]
                if isinstance(indexedOn, RtlSignalBase):
                    # [TODO] multidimensional indexing
                    return indexedOn, [d.ops[1]]
                else:
                    raise Exception("can not drive static value %s" % repr(indexedOn))
                
        except MultipleDriversExc:
            pass
        
    
    def __pow__(self, source):
        """
        Create assignment to this signal
        @attention: it is not power operator it is assignment
        @return: list of assignments
        """
        if isinstance(source, InterfaceBase):
            source = source._sig
        
        if source is None:
            source = self._dtype.fromPy(None)
        else:
            source = toHVal(source)
            source = source._dtype.convert(source, self._dtype)
        
 
        tmp = self._getIndexCascade()
        if tmp:
            mainSig, indexCascade = tmp
            self = mainSig
        else:
            indexCascade = None
            
        # self = self._tryMyIndexToEndpoint()
        a = Assignment(source, self, indexCascade)
        
        self.drivers.append(a)
        if not isinstance(source, Value):
            source.endpoints.append(a)
        
        return [a]



from hwt.hdlObjects.types.hdlType import HdlType 

class Boolean(HdlType):
    def __init__(self):
        super().__init__()
        self.name = 'boolean'
    
    @classmethod
    def getConvertor(cls):
        from hwt.hdlObjects.types.booleanConversions import convertBoolean
        return convertBoolean

    def bit_length(self):
        return 1

    def all_mask(self):
        return 1

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.booleanVal import BooleanVal 
            cls._valCls = BooleanVal
            return cls._valCls


from hwt.synthesizer.interfaceLevel.interface import Interface
from hwt.synthesizer.interfaceLevel.unit import Unit
from hwt.synthesizer.interfaceLevel.emptyUnit import EmptyUnit
from hwt.synthesizer.shortcuts import synthesised
from hwt.synthesizer.param import Param
from hwt.code import connect, c, If, In, Switch, FsmBuilder


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import shutil


from hwt.hdlObjects.architecture import Architecture
from hwt.hdlObjects.entity import Entity

from hwt.serializer.exceptions import SerializerException
from hwt.serializer.vhdl.serializer import VhdlSerializer
from hwt.synthesizer.interfaceLevel.unit import Unit
from hwt.synthesizer.uniqList import UniqList

def toRtl(unitOrCls, name=None, serializer=VhdlSerializer):
    """
    convert unit to rtl string using specified serializer
    """
    if not isinstance(unitOrCls, Unit):
        u = unitOrCls()
    else:
        u = unitOrCls
    
    u._loadDeclarations()
    if name is not None:
        u._name = name
    
    globScope = serializer.getBaseNameScope()
    codeBuff = []
    mouduleScopes = {}

    # unitCls : unitobj
    serializedClasses = {}
    
    # (unitCls, paramsValues) : unitObj
    # where paramsValues are dict name:value
    serializedConfiguredUnits = {}
    
    doSerialize = True
    for obj in u._toRtl():
        doSerialize = serializer.serializationDecision(obj, serializedClasses, serializedConfiguredUnits)
        if doSerialize:
            if isinstance(obj, Entity):
                s = globScope.fork(1)
                s.setLevel(2)
                mouduleScopes[obj] = s
                sc = serializer.Entity(obj, s)
            elif isinstance(obj, Architecture):
                try:
                    s = mouduleScopes[obj.entity]
                except KeyError:
                    raise SerializerException("Entity should be serialized before architecture of %s" % 
                                              (obj.getEntityName()))
                sc = serializer.Architecture(obj, s)
            else:
                sc = serializer.asHdl(obj)
        
            codeBuff.append(sc)
        else:
            try:
                name = "(" + obj.name + ")"
            except AttributeError:
                name = ""
                
            codeBuff.append(serializer.comment("Object of class %s%s was not serialized due its serializer mode" % (obj.__class__.__name__, name)))
    
    return serializer.formater(
                     "\n".join(codeBuff)
                     )

def synthesised(u):
    assert not u._wasSynthetised()
    if not hasattr(u, "_interfaces"):
        u._loadDeclarations()

    for _ in u._toRtl():
        pass
    return u

# [TODO] merge toRtlAndSave and toRtl
def toRtlAndSave(unit, folderName='.', name=None, serializer=VhdlSerializer):
    unit._loadDeclarations()
    os.makedirs(folderName, exist_ok=True)
    files = UniqList()
    if name is not None:
        unit._name = name
        
    globScope = serializer.getBaseNameScope()
    mouduleScopes = {}
    
    # unitCls : unitobj
    serializedClasses = {}
    
    # (unitCls, paramsValues) : unitObj
    # where paramsValues are dict name:value
    serializedConfiguredUnits = {}
    
    doSerialize = True
    for obj in unit._toRtl():
        doSerialize = serializer.serializationDecision(obj, serializedClasses, serializedConfiguredUnits)
        if doSerialize:
            if isinstance(obj, Entity):
                # we need to serialize before we take name, before name can change
                s = globScope.fork(1)
                s.setLevel(2)
                mouduleScopes[obj] = s
                
                sc = serializer.Entity(obj, s)
                fName = obj.name + serializer.fileExtension
                fileMode = 'w'
                
            elif isinstance(obj, Architecture):
                try:
                    s = mouduleScopes[obj.entity]
                except KeyError:
                    raise SerializerException("Entity should be serialized before architecture of %s" % 
                                              (obj.getEntityName()))
                sc = serializer.Architecture(obj, s)
                fName = obj.getEntityName() + serializer.fileExtension
                fileMode = 'a'
            else:
                if hasattr(obj, "_hdlSources"):
                    fName = None
                    for fn in obj._hdlSources:
                        if isinstance(fn, str):
                            shutil.copy2(fn, folderName)
                            files.append(fn)
                else:
                    sc = serializer.asHdl(obj)   
    
            if fName is not None:
                fp = os.path.join(folderName, fName)
                files.append(fp)
                
                with open(fp, fileMode) as f:
                    if fileMode == 'a':
                        f.write("\n")
                    f.write(
                        serializer.formater(sc)
                        )
    return files


def serializeAsIpcore(unit, folderName=".", name=None, serializer=VhdlSerializer):
    from hwt.serializer.ip_packager.packager import Packager
    p = Packager(unit, name=name)
    p.createPackage(folderName)
    return p
    
    



from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.constants import INTF_DIRECTION
from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import walkPhysInterfaces
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from hwt.synthesizer.vectorUtils import getWidthExpr
from hwt.pyUtils.arrayQuery import arr_any
from hwt.synthesizer.exceptions import IntfLvlConfErr


def splitToTermSet(width):
    """
    try split width expression to multiplicands
    """
    try:
        width = width.singleDriver()
    except (AttributeError, MultipleDriversExc):
        return set([width])
    if width.operator == AllOps.DIV:
        pass
    assert width.operator == AllOps.MUL
    return set(width.ops)


class InterfaceArray():
    """
    Interface with multiplied widths of signals to create array of interfaces
    """
    def __init__(self):
        self._arrayElemCache = []
        
    def _setMultipliedBy(self, factor, updateTypes=True):
        self._multipliedBy = factor
        for i in self._interfaces:
            i._setMultipliedBy(factor, updateTypes=updateTypes)
    
    def _tryExtractMultiplicationFactor(self):
        widths = []
        # collect all widths    
        for i in walkPhysInterfaces(self):
            if i._dtypeMatch or i._boundedEntityPort._dtype.constrain is None:
                # if is not constrained vector or type was resolved this can not be a interfaceArray
                return 
            w = getWidthExpr(i._boundedEntityPort._dtype)
            widths.append(w)
            
        # find what have all widths in common
        splitedWidths = map(splitToTermSet, widths)
        arrLen = None
        for w in splitedWidths:
            if arrLen is None:
                arrLen = w
            else:
                arrLen = arrLen.intersection(w)
        
        if len(arrLen) == 1:  # if is possible to determine arrLen
            return list(arrLen)[0] 
    
    def __len__(self):
        return self._multipliedBy
    
    def _initArrayItems(self):
        self._arrayElemCache = []
        for index in range(self._multipliedBy.staticEval().val):
            e = self._mkElemItem()
            e._name = "%d" % index
            e._parent = self
            e._isExtern = self._isExtern
            self._arrayElemCache.append(e)
    
    def _connectMyElems(self):
        if self._arrayElemCache: 
            for indx, e in enumerate(self._arrayElemCache):
                elemHasConnections = arr_any(walkPhysInterfaces(e),
                                             lambda x: bool(x._sig.endpoints) 
                                                       or bool(x._sig.drivers))
                if elemHasConnections:
                    e._resolveDirections()
                    
                    if e._direction == INTF_DIRECTION.MASTER:
                        e._connectTo(self, masterIndex=indx)
                    else:
                        self._connectTo(e, slaveIndex=indx)
        
    
    def __getitem__(self, key):
        if self._multipliedBy is None:
            raise IntfLvlConfErr("interface %s is not array and can not be indexe on" % self._name)
        return self._arrayElemCache[key]
    
    def _mkElemItem(self):
        e = self.__class__()
        e._updateParamsFrom(self)
        e._loadDeclarations()
        return e 
        



from hwt.hdlObjects.constants import INTF_DIRECTION
from hwt.interfaces.utils import cloneIntf
from hwt.synthesizer.interfaceLevel.unit import Unit
from hwt.synthesizer.param import evalParam, Param


class IpCoreWrapper(Unit):
    """
    Class which creates wrapper which converts all incompatible parts of unit to
    ipcore compatible this means:
    * convert array interface to multiple interfaces
    
    original unit will be placed inside as subunit named baseUnit
    
    """
    def __init__(self, baseUnit):
        super(IpCoreWrapper, self).__init__()
        self._baseUnit = baseUnit
    
    def _copyParamsAndInterfaces(self):
        for p in self._baseUnit._params:
            myP = Param(evalParam(p))
            self._registerParameter(p.name, myP)
            object.__setattr__(self, myP.name, myP)
        
        origToWrapInfMap = {}
        
        for intf in self.baseUnit._interfaces:
            if intf._multipliedBy is not None:
                for i in range(evalParam(intf._multipliedBy).val):
                    myIntf = cloneIntf(intf)
                    name = intf._name + "_%d" % i 
                    
                    self._registerInterface(name, myIntf)
                    object.__setattr__(self, name, myIntf)
                    
                    try:
                        ia = origToWrapInfMap[intf]
                    except KeyError:
                        ia = []
                        origToWrapInfMap[intf] = ia
                    
                    ia.append(myIntf)
            
            else:
                myIntf = cloneIntf(intf)
                
                self._registerInterface(intf._name, myIntf)
                object.__setattr__(self, intf._name, myIntf)
                
                origToWrapInfMap[intf] = myIntf
            
        for i in self._interfaces:
            self._loadInterface(i, True)   
        
        return origToWrapInfMap   
    
    
    def _getDefaultName(self):
        return self._baseUnit.__class__.__name__
          
    def _lazyLoadParamsAndInterfaces(self):
        self._cntx.globals = self._globalsFromParams()
        
        # prepare signals for interfaces     
        for i in self._interfaces:
            assert i._isExtern
            signals = i._signalsForInterface(self._cntx)
            self._externInterf.extend(signals)
    
    def _connectBaseUnitToThisWrap(self, origToWrapInfMap):
        for baseIntf, wrapIntf in origToWrapInfMap.items():
            if baseIntf._direction is INTF_DIRECTION.MASTER:
                if isinstance(wrapIntf, list):
                    for i, _wrapIntf in enumerate(wrapIntf):
                        _wrapIntf ** baseIntf[i]
                else:
                    wrapIntf ** baseIntf
            else:
                if isinstance(wrapIntf, list):
                    for i, _wrapIntf in enumerate(wrapIntf):
                        baseIntf[i] ** _wrapIntf
                else:
                    baseIntf ** wrapIntf
    
    def _impl(self):
        self.baseUnit = self._baseUnit
        origToWrapInfMap = self._copyParamsAndInterfaces()
        self._lazyLoadParamsAndInterfaces()
        self._connectBaseUnitToThisWrap(origToWrapInfMap)



from hwt.hdlObjects.types.hdlType import HdlType 

class Array(HdlType):
    """
    vldMask and eventMask on Array_val instance is not used instead of that
    these flags on elements are used
    """
    def __init__(self, elmType, size):
        super(Array, self).__init__()
        self.elmType = elmType
        self.size = size
    
    def __hash__(self):
        return hash((self.elmType, self.size))
    
    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.arrayVal import ArrayVal 
            cls._valCls = ArrayVal
            return cls._valCls


from copy import copy

from hwt.hdlObjects.constants import DIRECTION, INTF_DIRECTION
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwt.synthesizer.interfaceLevel.interfaceUtils.directionFns import InterfaceDirectionFns
from hwt.synthesizer.interfaceLevel.interfaceUtils.hdlExtraction import ExtractableInterface
from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import NotSpecified
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.interfaceLevel.propDeclrCollector import PropDeclrCollector
from hwt.synthesizer.param import Param
from hwt.synthesizer.rtlLevel.netlist import RtlNetlist
from hwt.synthesizer.vectorUtils import fitTo, aplyIndexOnSignal


def _defaultUpdater(self, onParentName, p):
    self._replaceParam(onParentName, p)


class Interface(InterfaceBase, ExtractableInterface, PropDeclrCollector, InterfaceDirectionFns):
    """
    Base class for all interfaces in interface synthesizer

    @cvar _NAME_SEPARATOR: separator for nested interface names
    @ivar _params: [] of parameter
    @ivar _interfaces: [] sub interfaces
    @ivar _alternativeNames: [] of alternative names
    @ivar _name: name assigned during synthesis
    @ivar _parent: parent object (Unit or Interface instance)
    @ivar _isExtern: If true synthesizer sets it as external port of unit

    #only interfaces without _interfaces have:
    @ivar _sig: rtl level signal instance
    @ival _sigInside : _sig after toRtl conversion is made (after toRtl conversion
                    _sig is signal for parent unit and _sigInside is signal
                    in original unit, this separates process of translating units)
    @ivar _boundedEntityPort: entityPort for which was this interface created
    @ivar _boundedSigLvlUnit: RTL unit for which was this interface created


    Agenda of direction:
    @ivar _masterDir: specifies which direction has this interface at master
    @ivar _direction: means actual direction of this interface resolved by its drivers
    @ivar _cntx: rtl netlist context of all signals and params on this interface
                 after interface is registered on parent _cntx is merged

    Agenda of simulations:
    @ivar _ag: agent object connected to this interface (initialized by simulator)
    """

    _NAME_SEPARATOR = "_"

    def __init__(self, masterDir=DIRECTION.OUT, multipliedBy=None, alternativeNames=None, loadConfig=True):
        """
        This constructor is called when constructing new interface, it is usually done
        manually while creating Unit or
        automatically while extracting interfaces from UnitWithSoure
        @param masterDir: direction which this interface should have for master
        @param multiplyedBy: this can be instance of integer or Param, this mean the interface
                         is array of the interfaces where multiplyedBy is the size
        @param alternativeNames: alternative names which are used for interface extraction from hdl
                                [TODO] remove
        @param loadConfig: do load config in __init__
        """
        self._setAttrListener = None
        super().__init__()
        if multipliedBy is not None:
            multipliedBy = toHVal(multipliedBy)
        self._multipliedBy = multipliedBy
        self._masterDir = masterDir
        self._direction = INTF_DIRECTION.UNKNOWN

        # resolve alternative names
        if not alternativeNames:
            if hasattr(self.__class__, "_alternativeNames"):
                self._alternativeNames = copy(self.__class__._alternativeNames)
            else:
                self._alternativeNames = []
        else:
            self._alternativeNames = alternativeNames

        # set default name to this interface
        if not hasattr(self, "_name"):
            if self._alternativeNames:
                self._name = self._alternativeNames[0]
            else:
                self._name = ''

        self._cntx = RtlNetlist(self)

        if loadConfig:
            self._loadConfig()

        # flags for better design error detection
        self._isExtern = False
        self._isAccessible = True
        self._dirLocked = False
        self._ag = None

    def _loadDeclarations(self):
        if not hasattr(self, "_interfaces"):
            self._interfaces = []
        self._setAttrListener = self._declrCollector
        self._declr()
        self._setAttrListener = None
        for i in self._interfaces:
            # inherit _multipliedBy and update dtype on physical interfaces
            if i._multipliedBy is None:
                i._multipliedBy = self._multipliedBy
            i._isExtern = self._isExtern
            i._loadDeclarations()

            # apply multiplier at dtype of signals
            if i._multipliedBy is not None:
                if not i._interfaces:
                    i._injectMultiplerToDtype()
                    i._initArrayItems()

    def _clean(self, rmConnetions=True, lockNonExternal=True):
        """Remove all signals from this interface (used after unit is synthesized
         and its parent is connecting its interface to this unit)"""
        if self._interfaces:
            for i in self._interfaces:
                i._clean(rmConnetions=rmConnetions, lockNonExternal=lockNonExternal)
        else:
            self._sigInside = self._sig
            del self._sig

        self._dirLocked = False
        if lockNonExternal and not self._isExtern:
            self._isAccessible = False  # [TODO] mv to signal lock
        for e in self._arrayElemCache:
            e._clean(rmConnetions=rmConnetions, lockNonExternal=lockNonExternal)

    def _connectToIter(self, master, masterIndex=None, slaveIndex=None, exclude=None, fit=False):
        if exclude and (self in exclude or master in exclude):
            return

        if self._interfaces:
            for ifc in self._interfaces:
                if exclude and ifc in exclude:
                    continue
                mIfc = getattr(master, ifc._name)

                if exclude and mIfc in exclude:
                    continue

                if mIfc._masterDir == DIRECTION.OUT:
                    if ifc._masterDir != mIfc._masterDir:
                        raise IntfLvlConfErr("Invalid connection %s <= %s" % (repr(ifc), repr(mIfc)))

                    yield from ifc._connectTo(mIfc, masterIndex=masterIndex, slaveIndex=slaveIndex,
                                                    exclude=exclude, fit=fit)
                else:
                    if ifc._masterDir != mIfc._masterDir:
                        raise IntfLvlConfErr("Invalid connection %s <= %s" % (repr(mIfc), repr(ifc)))

                    yield from mIfc._connectTo(ifc, masterIndex=slaveIndex, slaveIndex=masterIndex,
                                                    exclude=exclude, fit=fit)
        else:
            dstSig = toHVal(self)
            srcSig = toHVal(master)

            if masterIndex is not None:
                srcSig = aplyIndexOnSignal(srcSig, dstSig._dtype, masterIndex)

            if slaveIndex is not None:
                dstSig = aplyIndexOnSignal(dstSig, srcSig._dtype, slaveIndex)

            if fit:
                srcSig = fitTo(srcSig, dstSig)

            yield dstSig.__pow__(srcSig)

    def _connectTo(self, master, masterIndex=None, slaveIndex=None, exclude=None, fit=False):
        """
        connect to another interface interface (on rtl level)
        works like self <= master in VHDL
        """
        return list(self._connectToIter(master, masterIndex, slaveIndex,
                                        exclude, fit))

    def __pow__(self, other):
        """
        @attention: ** operator is used as "assignment" it creates connection between interface and other
        """
        return self._connectTo(other)

    def _signalsForInterface(self, context, prefix='', typeTransform=None):
        """
        generate _sig for each interface which has no subinterface
        if already has _sig return it instead
        @param typeTransform: optional function (type) returns modified type for signal
        """
        sigs = []
        if self._interfaces:
            for intf in self._interfaces:
                sigs.extend(intf._signalsForInterface(context, prefix,
                                                      typeTransform=typeTransform))
        else:
            if hasattr(self, '_sig'):
                sigs = [self._sig]
            else:
                t = self._dtype
                if typeTransform is not None:
                    t = typeTransform(t)

                s = context.sig(prefix + self._getPhysicalName(), t)
                s._interface = self
                self._sig = s

                if hasattr(self, '_boundedEntityPort'):
                    self._boundedEntityPort.connectSig(self._sig)
                sigs = [s]

        if self._multipliedBy is not None:
            for elemIntf in self._arrayElemCache:
                # elemPrefix = prefix + self._NAME_SEPARATOR + elemIntf._name
                elemIntf._signalsForInterface(context, prefix, typeTransform=typeTransform)
                # they are not in sigs because they are not main signals

        return sigs

    def _getPhysicalName(self):
        """Get name in HDL """
        if hasattr(self, "_boundedEntityPort"):
            return self._boundedEntityPort.name
        else:
            return self._getFullName().replace('.', self._NAME_SEPARATOR)

    def _getFullName(self):
        """get all name hierarchy separated by '.' """
        name = ""
        tmp = self
        while isinstance(tmp, Interface):
            if hasattr(tmp, "_name"):
                n = tmp._name
            else:
                n = ''
            if name == '':
                name = n
            else:
                name = n + '.' + name
            if hasattr(tmp, "_parent"):
                tmp = tmp._parent
            else:
                tmp = None
        return name

    def _replaceParam(self, pName, newP):
        """Replace parameter in configuration stage"""
        p = getattr(self, pName)
        i = self._params.index(p)
        assert i > -1
        self._params[i] = newP
        del p._scopes[self]  # remove reference from old param
        newP._registerScope(pName, self)
        setattr(self, pName, newP)

    def _updateParamsFrom(self, otherObj, updater=_defaultUpdater, exclude=None):
        """
        update all parameters which are defined on self from otherObj
        @param exclude: iterable of parameter on other object which should be excluded
        """
        excluded = set()
        if exclude is not None:
            exclude = set(exclude)
        
        for parentP in otherObj._params:
            if exclude and parentP in exclude:
                excluded.add(parentP)
                continue
            try:
                _, onParentName = parentP._scopes[otherObj]
            except KeyError as e:
                raise e
            try:
                myP = getattr(self, onParentName)
                if not isinstance(myP, Param):
                    raise AttributeError()
            except AttributeError:
                continue

            updater(self, onParentName, parentP)
        
        if exclude is not None:
            assert excluded == exclude

    def _getIpCoreIntfClass(self):
        raise NotSpecified()

    def _getSimAgent(self):
        raise NotSpecified()

    def __repr__(self):
        s = [self.__class__.__name__]
        s.append("name=%s" % self._getFullName())
        if hasattr(self, '_width'):
            s.append("_width=%s" % str(self._width))
        if hasattr(self, '_masterDir'):
            s.append("_masterDir=%s" % str(self._masterDir))
        return "<%s>" % (', '.join(s))



from hwt.simulator.agentBase import SyncAgentBase


class RdSyncedAgent(SyncAgentBase):
    """
    Simulation/verification agent for RdSynced interface
    """
    def __init__(self, intf, clk=None, rstn=None):
        super().__init__(intf, clk=clk, rstn=rstn, allowNoReset=True)
        self.actualData = None
        
    def monitor(self, s):
        """Collect data from interface"""
        intf = self.intf
        
        if self.notReset(s) and self.enable:
            s.w(1, intf.rd)
            
            yield s.updateComplete
            
            d = self.doRead(s)
            self.data.append(d)
        else:
            s.w(0, intf.rd)
    
    def doRead(self, s):
        """extract data from interface"""
        return s.read(self.intf.data)
        
    def doWrite(self, s, data):
        """write data to interface"""
        s.w(data, self.intf.data)
        
    def driver(self, s):
        """Push data to interface"""
        intf = self.intf
        
        if self.actualData is None and self.data:
            self.actualData = self.data.pop(0)
        
        self.doWrite(s, self.actualData)
        if self.notReset(s) and self.actualData is not None and self.enable:
            s.w(1, intf.vld)
            yield s.updateComplete
                    
            d = self.doRead(s)
            self.data.append(d)
        else:
            s.w(0, intf.vld)



    


from hwt.serializer.vhdl.serializer import VhdlSerializer


class VivadoTclExpressionSerializer(VhdlSerializer):
    @staticmethod
    def SignalItem(si, declaration=False):
        assert(declaration == False)
        if si.hidden:
            return VhdlSerializer.asHdl(si.origin)
        else:
            return "spirit:decode(id('MODELPARAM_VALUE.%s'))" % (si.name)




from hwt.interfaces.std import Clk, Rst_n, Rst
from hwt.code import connect

def addClkRstn(self):
    self.clk = Clk()
    self.rst_n = Rst_n()

def addClkRst(self):
    self.clk = Clk()
    self.rst = Rst()


def _tryConnect(src, unit, intfName):
    try:
        dst = getattr(unit, intfName)
    except AttributeError:
            dst = None
    if dst is not None:
        connect(src, dst)

def propagateClk(self):
    clk = self.clk
    for u in self._units:
        _tryConnect(clk, u, 'clk')
    
def propagateClkRstn(self):
    clk = self.clk
    rst_n = self.rst_n
    
    for u in self._units:
        _tryConnect(clk, u, 'clk')
        _tryConnect(rst_n, u, 'rst_n')
        _tryConnect(~rst_n, u, 'rst')
        
def propagateClkRst(self):
    clk = self.clk
    rst = self.rst
    
    for u in self._units:
        _tryConnect(clk, u, 'clk')
        _tryConnect(~rst, u, 'rst_n')
        _tryConnect(rst, u, 'rst')

def cloneIntf(intf):
    i = intf.__class__()
    i._updateParamsFrom(intf)
    return i





from hwt.hdlObjects.types.defs import INT, STR, BOOL
from hwt.hdlObjects.value import Value
from hwt.hdlObjects.variables import SignalItem
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase


defaultConversions = {int: INT,
                      str: STR,
                      bool: BOOL}
def toHVal(op):
    """Convert python value object to object of hdl type value"""
    if isinstance(op, Value) or isinstance(op, SignalItem):
        return op
    elif isinstance(op, InterfaceBase):
        return op._sig
    else:
        try:
            hType = defaultConversions[type(op)]
        except KeyError:
            hType = None
        
        if hType is None:
            raise TypeError("%s" % (op.__class__))
        return  hType.fromPy(op)
    
def checkOperands(ops):
    _ops = []
    for op in ops:
        _ops.append(toHVal(op))
    return _ops



from simpy.events import NORMAL

from hwt.hdlObjects.value import Value
from hwt.simulator.hdlSimConfig import HdlSimConfig
from hwt.simulator.simModel import mkUpdater, mkArrayUpdater
from hwt.simulator.simulatorCore import HdlEnvironmentCore
from hwt.simulator.utils import valueHasChanged
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase


def isEvDependentOn(sig, process):
    if sig is None:
        return False
    return process in sig.simFallingSensProcs or process in sig.simRisingSensProcs
    

class HdlSimulator(HdlEnvironmentCore):
    """
    [HOW IT WORKS]
    
    Every signal should have single or none driver
    -> (no conflict resolving is needed (is already done by synthesizer))
    
    Every signal is initialized at start with its default value
    -> (no drove, constant drove solved)
    
    Every assignment has it's delay, clock dependent assignments has delay != 0
    (they should be marked before sim started)
    -> (memories and register solved)
    
    Every signal value changing object (assignment, index...) should on simEval() yield
       tuple (updateDelayTime, applicator) where applicator is function(oldValue)
       and returns tuple (newValue, valueHasChangedFlag) 
    -> (updating value in arrays, structs etc. solved)
    
    Every interprocess signal is marked by synthesizer and it can not be directly updated
       by any process, process should only return tuple (updateDelayTime, applicator)
       and let simulator to update it for others, any other signals are evaluated as expression
       by every process
    every process drives only one signal
    every process uses sensitivity-list like in other languages (but it is generated automatically)
    -> (communication between process solved)
    
    Hdl processes can not contain any wait statements etc. only simulation processes can.
    Simulation processes are written in python.
    -> (using hdl as main simulator driver is not efficient and thats why it is not supported
        and it is easy to just read hdl process with unsupported statements and translate them to 
        simulator commands)
    
    
    HWprocesses have smaller priority than simulation processes
    this allows simplyfi logic of all agents
    when simulation process is executed HW part did not anything in this time
    so simulation process can prepare anythink for HW part (= can write)
    if simulation process need to read it has to yield simulator.updateComplete
    first, process then be waken after reaction of HW in this time
    -> agents are greatly simplified, they just need to yield simulator.updateComplete
       before read 
       Do not read before write in single time (potential combinational loop), because
       of event dependent HW processes will not be reevalueated 
       
    
    This simulation is made to check if hsl code behaves the same way as hdl. 
    It has some limitations from HDL point of view but every single one can be rewritten to supported format.
    Hdl synthesizer of HWToolkit does it automatically.
    
    
    @ivar updateComplete: this event is triggered when there are not any values to apply in this time
    @ivar valuesToApply: is container to for quantum of values which should be applied in single time 
    @ivar env: simpy enviromnment
    @ivar applyValuesPlaned: flag if there is planed applyValues for current values quantum
    @ivar seqProcsToRun: list of event dependent processes which should be evaluated after 
                                applyValEv
    """
    # time after values which are event dependent will be applied
    # this is random number smaller than any clock half-period
    EV_DEPENDENCY_SLOWDOWN = 500
    
    PRIORITY_APPLY_COMB = NORMAL + 1
    PRIORITY_APPLY_SEQ = PRIORITY_APPLY_COMB + 1 
     
    # http://heather.cs.ucdavis.edu/~matloff/156/PLN/DESimIntro.pdf
    def __init__(self, config=None):
        super(HdlSimulator, self).__init__()
        if config is None:
            config = HdlSimConfig() 
            
        self.config = config
        self.updateComplete = self.event()
        self.applyValEv = None
        self.runSeqProcessesEv = None
        
        # (signal, value) tupes which should be applied before new round of processes
        #  will be executed
        self.valuesToApply = []
        self.seqProcsToRun = []
        
    
    def addHwProcToRun(self, trigger, proc):
        # first process in time has to plan executing of apply values on the end of this time
        if self.applyValEv is None:
            # (apply on end of this time to minimalize process reevaluation)
            self.scheduleAplyValues()
            
        if isEvDependentOn(trigger, proc):
            if self.now == 0:
                pass  # pass event dependent on startup
            self.seqProcsToRun.append(proc)
            
        else:
            actionSet = set(proc(self))
            res = self.conflictResolvStrategy(actionSet)
            if res:
                dst, updater, isEvDependent = res
                if trigger is not None:
                    assert not isEvDependent, "trigger %r, proc %r" % (trigger, proc)
                self.valuesToApply.append((dst, updater, isEvDependent, proc))
            
        
    def _initUnitSignals(self, unit):
        """
        Inject default values to simulation
        @return: generator of all HWprocess 
        """
        for s in unit._cntx.signals:
            v = s.defaultVal.clone()
            
            # force update all signals to deafut values and propagate it    
            s.simUpdateVal(self, mkUpdater(v, False))
            
        for u in unit._units:
            self._initUnitSignals(u)
        
        for p in unit._processes:
            self.addHwProcToRun(None, p) 
    
    def scheduleAplyValues(self):
        applyVal = self.applyValEv = self.event()
        applyVal._ok = True
        applyVal._value = None
        applyVal.callbacks.append(self.applyValues)
        
        self.schedule(applyVal, priority=self.PRIORITY_APPLY_COMB)
        
        if self.runSeqProcessesEv is not None:
            return
        
        #print(self.now, "sched")
        assert not self.seqProcsToRun       
        runSeq = self.runSeqProcessesEv = self.event()
        runSeq._ok = True
        runSeq._value = None
        runSeq.callbacks.append(self.runSeqProcesses)
        
        self.schedule(runSeq, priority=self.PRIORITY_APPLY_SEQ)
    
    def conflictResolvStrategy(self, actionSet):
        """
        This functions resolves 
        @param actionSet: set of actions made by process
        
        """
        invalidate = False
        l = len(actionSet)
        if l == 0:
            return
        elif l == 1:
            res = actionSet.pop()
        else:
            # we are driving signal with two different values so we invalidate result
            res = list(actionSet.pop())
            invalidate = True

        l = len(res)
        if l == 4:
            dst, val, indexes, isEvDependent = res
            return (dst, mkArrayUpdater(val, indexes, invalidate), isEvDependent)
        else:
            dst, val, isEvDependent = res
                
            # print(self.now, dst, val)
            return (dst, mkUpdater(val, invalidate), isEvDependent)
    
    
    def runSeqProcesses(self, ev):
        #print(self.now, "run")
        updates = []
        for proc in self.seqProcsToRun:
            actionSet = set(proc(self))
            if actionSet:
                v = self.conflictResolvStrategy(actionSet)
                updates.append(v)
                
        self.seqProcsToRun = []
        self.runSeqProcessesEv = None
        for s, updater, _ in updates:
            s.simUpdateVal(self, updater)
        
    def applyValues(self, ev):
        va = self.valuesToApply
        
        # log if there are items to log
        if va and self.config.logApplyingValues:
            self.config.logApplyingValues(self, va)
            
        self.valuesToApply = []

        # apply values to signals, values can overwrite each other
        # but each signal should be driven by only one process and
        # it should resolve value collision
        for s, vUpdater, isEventDependent, comesFrom in va:
            if isEventDependent:
                self.seqProcsToRun.append(comesFrom)
            else:
                s.simUpdateVal(self, vUpdater)
            
            
        # processes triggered from simUpdateVal can add nev values
        if self.valuesToApply:
            self.scheduleAplyValues()
            return
        
        # activate updateComplete if this was last applyValues() in this time        
        self.updateComplete.succeed()  # trigger
        self.updateComplete = self.event()  # regenerate event
        self.applyValEv = None 
           
    def read(self, sig):
        """
        Read value from signal or interface
        """
        if isinstance(sig, InterfaceBase):
            sig = sig._sigInside
        return sig._val.clone()
    
    def write(self, val, sig):
        """
        Write value to signal or interface.
        """
        if isinstance(val, Value):
            v = val.clone()
        else:
            v = sig._dtype.fromPy(val)
        
        v.updateTime = self.now
        if isinstance(sig, InterfaceBase):
            sig = sig._sigInside

        v = v._convert(sig._dtype)
        
        sig.simUpdateVal(self, lambda curentV: (valueHasChanged(curentV, v), v))
        
        if not sig.simSensProcs and self.applyValEv is not None:
            # in some cases simulation process can wait on all values applied
            # signal value was changed but there are no sensitive processes to it
            # because of this applyValues is never planed and but should be
            self.scheduleAplyValues()
            
    def wait(self, time):
        return self.timeout(time)
    
    def simUnit(self, synthesisedUnit, time, extraProcesses=[]):
        """
        Run simulation
        """
        self.config.beforeSim(self, synthesisedUnit)
        
        for p in extraProcesses:
            self.process(p(self))
        
        self._initUnitSignals(synthesisedUnit)
       
        self.run(until=time)
    
    # shortcuts
    r = read    
    w = write
        



from hwt.hdlObjects.assignment import Assignment
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps, isEventDependentOp
from hwt.hdlObjects.portItem import PortItem
from hwt.hdlObjects.constants import DIRECTION, SENSITIVITY
from hwt.hdlObjects.statements import IfContainer, SwitchContainer
from hwt.hdlObjects.value import Value
from hwt.pyUtils.arrayQuery import where
from hwt.synthesizer.param import Param
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc


def signalHasDriver(sig):
    for _ in walkSignalDrivers(sig):
        return True
    return False

def walkSignalDrivers(sig):
    def assign2Me(ep):
        if isinstance(ep, Assignment):
            return True
        elif isinstance(ep, PortItem) and ep.dst is sig: 
            return True
        else:
            return None
            
    return where(sig.drivers, assign2Me)
 
def walkSigExpr(sig):
    """
    Walk any object connected to this signal 
    """
    yield from sig.drivers
    yield from sig.endpoints

def walkUnitInputPorts(unit):
    for portItem in unit.ports:
        if portItem.direction == DIRECTION.IN:
            yield portItem


def walkAllOriginSignals(sig, discovered=None):
    """
    Walk every signal which has no driver and is used as driver of this signal.
    Goal is walk every generic in static expr. evaluation.
    """
    if discovered is None:
        discovered = set()
    if isinstance(sig, Value):
        raise StopIteration()
    if not isinstance(sig, RtlSignalBase):
        raise  AssertionError("Expected only instances of signal, got: %s" % 
                              (repr(sig)))
    if sig in discovered:
        return
    discovered.add(sig)
    
    if sig.drivers:
        for obj in sig.drivers:
            if isinstance(obj, Value):
                raise StopIteration()
            elif isinstance(obj, Operator):
                for op in obj.ops:
                    if isinstance(op, RtlSignalBase):
                        yield from walkAllOriginSignals(op, discovered=discovered)
            elif isinstance(obj, RtlSignalBase):
                yield from walkAllOriginSignals(obj, discovered)
            elif isinstance(obj, Assignment):
                yield from walkAllOriginSignals(obj.src, discovered)
            else:
                raise TypeError("walkAllOriginSignals not implemented for %s" % (str(obj)))
    else:
        yield sig

def walkSignalsInExpr(expr):
    if isinstance(expr, Value):
        return
    elif isinstance(expr, Operator):
        for op in expr.ops:
            if op is not expr:
                yield from walkSignalsInExpr(op)
    elif isinstance(expr, RtlSignalBase):
        if hasattr(expr, "origin"):
            yield from  walkSignalsInExpr(expr.origin)
        else:
            yield expr
    else:
        raise Exception("Unknown node '%s' type %s" % 
                        (repr(expr), str(expr.__class__)))

class EventDependencyReached(Exception):
    """
    Exception which signalize that walkDriversInExpr 
    has reached event dependent operator
    """
    def __init__(self, evOp):
        self.evOp = evOp

def walkDriversInExpr(expr, seenSet):
    """
    @return: generators of RtlSignal
            where signal is in expression and is not used in event dependent expression
    @raise EventDependencyReached: when this generator steps on event dependent operator 
    @attention: if event operator is found in expression, only sensitivity EventDependencyReached is raised
                this may case some synthesis (Vivado, ISE, Quartus ...) to complain about sensitivity, 
                but it will work 
    """
    if isinstance(expr, (Value, Param)):
        pass
    elif isinstance(expr, RtlSignalBase):
        if expr in seenSet:
            pass
        else:
            seenSet.add(expr)
            
            if not expr.hidden:
                yield expr
                return
            
            try:
                op = expr.singleDriver()
            except MultipleDriversExc:
                yield expr
                return
            
            if not isinstance(op, Operator):
                yield expr
                return
            
            if isEventDependentOp(op.operator):
                raise EventDependencyReached(op)
            else:
                for operand in op.ops:
                    yield from walkDriversInExpr(operand, seenSet)
                    
    else:           
        raise TypeError(expr)

def discoverEventDependency(sig):
    """
    walk signals drivers and yields whose signals which are in some event operator
    """
    try:
        drivers = sig.drivers
    except AttributeError:
        return
    
    if len(drivers) == 1:
        d = drivers[0]
        if isinstance(d, Operator):
            if isEventDependentOp(d.operator):
                yield d.ops[0]
            else:
                for op in d.ops:
                    yield from discoverEventDependency(op)


def isEdgeDependent(sens):
    return sens == SENSITIVITY.RISING or sens == SENSITIVITY.FALLING

def _discoverSensitivity(statement, seenSet, isTop):
    try:
        if isinstance(statement, Assignment):
            if statement.indexes:
                for i in statement.indexes:
                    yield from walkDriversInExpr(i, seenSet)
            yield from walkDriversInExpr(statement.src, seenSet)
        elif isinstance(statement, IfContainer):
            # if true
            for c in statement.cond:
                yield from walkDriversInExpr(c, seenSet)
            for stm in statement.ifTrue:
                yield from _discoverSensitivity(stm, seenSet, False)
            # elifs
            for cond, stms in statement.elIfs:
                for c in cond:
                    yield from walkDriversInExpr(c, seenSet)
                for stm in stms:
                    yield from _discoverSensitivity(stm, seenSet, False)
            # else
            for stm in statement.ifFalse:
                yield from _discoverSensitivity(stm, seenSet, False)
            
        elif isinstance(statement, SwitchContainer):
            yield from walkDriversInExpr(statement.switchOn, seenSet)
            for cond, stms in statement.cases:
                # yield from walkDriversInExpr(cond, seenSet)
                for stm in stms:
                    yield from _discoverSensitivity(stm, seenSet, False)
        else:
            raise TypeError(statement)
        
    except EventDependencyReached as e:
        if isTop:
            yield e.evOp
        else:
            raise e

def discoverSensitivity(statement):
    """
    @return: generators of (sensitivity, signal)
            where sensitivity is member of SENSITIVITY enum
    """
    yield from _discoverSensitivity(statement, set(), True) 
        
# walks code, and returns assignments and port items which are driving this signal
def walkSigSouces(sig, parent=None):    
    if isinstance(sig, Operator):
        if sig.operator != AllOps.INDEX:  # [TODO] more test to assert this cond. will work
            for op in sig.ops:
                if not op is parent:
                    yield from walkSigSouces(op)
    elif isinstance(sig, RtlSignalBase):
        for e in sig.drivers:
            if isinstance(e, PortItem):
                if not e.unit.discovered:
                    yield e
            elif isinstance(e, Assignment) and not e.src is sig:
                yield e
            else:
                yield from walkSigSouces(e, sig)
    else:
        raise Exception("Cant walk node %s" % repr(sig))
        
        



from hwt.synthesizer.interfaceLevel.interface import Interface
from hwt.interfaces.std import Signal


class DifferentialSig(Interface):
    def _declr(self):
        self.n = Signal()
        self.p = Signal()



from hwt.hdlObjects.constants import INTF_DIRECTION
from hwt.synthesizer.param import evalParam

def autoAddAgents(unit, propName="_ag"):
    """
    Walk all interfaces on unit and instantiate actor for every interface.
    
    @return: all monitor/driver functions which should be added to simulation as processes
     
    """
    proc = []
    for intf in unit._interfaces:
        try:
            agentCls = intf._getSimAgent()
        except NotImplementedError:
            raise NotImplementedError(("Interface %s\n" + 
                            "has not any simulation agent class assigned") % (str(intf)))
        
        if intf._multipliedBy is not None:
            agentCnt = evalParam(intf._multipliedBy).val
            agent = []
            for i in range(agentCnt):
                a = agentCls(intf[i])
                agent.append(a)
                setattr(intf[i], propName, a)
        else:
            agent = agentCls(intf)
            setattr(intf, propName, agent)
        
        if intf._multipliedBy is None:
            agent = [agent, ]
            
        if intf._direction == INTF_DIRECTION.MASTER:
            agProcs = list(map(lambda a: a.getMonitors(), agent))
        elif intf._direction == INTF_DIRECTION.SLAVE:
            agProcs = list(map(lambda a: a.getDrivers(), agent))
        else:
            raise NotImplementedError("intf._direction %s" % str(intf._direction))
        
        for p in agProcs:
            proc.extend(p)
    
    return proc

def valuesToInts(values):
    """
    Iterable of values to ints (nonvalid = None)
    """
    res = []
    for d in values:
        res.append(valToInt(d))
    return res

def valToInt(v):
    if v.vldMask == v._dtype.all_mask():
        return v.val
    else:
        return None

def agInts(interface):
    """
    Convert all values which has agent collected in time >=0 to integer array.
    Invalid value will be None.
    """
    return valuesToInts(interface._ag.data)



from random import Random


def valueHasChanged(valA, valB):
    return valA.val is not valB.val or valA.vldMask != valB.vldMask


def agent_randomize(agent, timeQuantum, seed):
    random = Random(seed)

    def randomEnProc(simulator):
        # small space at start to modify agents when they are inactive
        yield simulator.wait(timeQuantum / 4)
        while True:
            agent.enable = random.random() < 0.5
            delay = int(random.random() * timeQuantum)
            yield simulator.wait(delay)

    return randomEnProc



from hwt.hdlObjects.types.hdlType import HdlType 


class Integer(HdlType):
    
    def __init__(self, _min=None, _max=None):
        super(Integer, self).__init__()
        self.min = _min
        self.max = _max
    
    def __eq__(self, other):
        return self is other or (
            type(self) == type(other) and self.max == other.max and self.min == other.min
                                 )
    def __hash__(self):
        return hash((self.max, self.min))
    
    def all_mask(self):
        return 1

    @classmethod
    def getConvertor(cls):
        from hwt.hdlObjects.types.integerConversions import convertInteger
        return convertInteger
    
    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.integerVal import IntegerVal 
            cls._valCls = IntegerVal
            return cls._valCls


from copy import deepcopy

from hwt.hdlObjects.function import Function
from hwt.hdlObjects.value import Value
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal, RtlSignalBase 


class InvalidOperandExc(Exception):
    pass

def getCtxFromOps(ops):
    for o in ops:
        if isinstance(o, RtlSignalBase):
            return o.ctx
    raise TypeError("Can not find context because there is no signal in ops")
    

class Operator():
    """
    Class of operator in expression tree
    @ivar ops: list of operands
    @ivar evalFn: function to evaluate this operator
    @ivar operator: OpDefinition instance 
    @ivar result: result signal of this operator
    """
    def __init__(self, operator, operands):
        self.ops = list(operands)
        self.operator = operator
        self.result = None
        self._isDriver = True
            
    def registerSignals(self, outputs=[]):
        """
        Register potential signals to drivers/endpoints
        """
        for o in self.ops:
            if isinstance(o, RtlSignalBase):
                if o in outputs:
                    o.drivers.append(self)
                else:
                    o.endpoints.append(self)
            elif isinstance(o, (Value, Function)):
                pass
            else:
                raise NotImplementedError("Operator operands can be only signal or values got:%s" % repr(o))

    def staticEval(self):
        """
        Recursively statistically evaluate result of this operator
        """
        for o in self.ops:
            o.staticEval()
        self.result._val = self.evalFn()
        
    def evalFn(self, simulator=None):
        """
        Syntax sugar
        """
        return self.operator.eval(self, simulator=simulator)
    
    def __eq__(self, other):
        return self is other or (
             type(self) == type(other) 
            and self.operator == other.operator \
            and self.ops == other.ops)
    
    @staticmethod
    def withRes(opDef, operands, resT, outputs=[]):
        """
        Create operator with result signal
        """
        op = Operator(opDef, operands)
        out = RtlSignal(getCtxFromOps(operands), None, resT)
        out.drivers.append(op)
        out.origin = op
        op.result = out
        op.registerSignals(outputs)
        
        return out
    
    def __deepcopy__(self, memo=None):
        try:
            return memo[self]
        except KeyError:
            o = Operator(None, [])
            memo[id(self)] = o
            for k, v in self.__dict__.items():
                setattr(o, k, deepcopy(v, memo))

            return o
                
    def __hash__(self):
        return hash((self.operator, frozenset(self.ops)))
    
        
    def __repr__(self):
        return "<%s operator:%s, ops:%s>" % (self.__class__.__name__,
                                             repr(self.operator), repr(self.ops))



from hwt.hdlObjects.constants import READ, WRITE
from hwt.simulator.agentBase import SyncAgentBase
from hwt.simulator.shortcuts import oscilate


class BramPort_withoutClkAgent(SyncAgentBase):
    """
    @ivar requests: list of tuples (request type, address) - used for driver
    @ivar data:     list of data in memory, used for monitor
    """
    def __init__(self, intf, clk=None, rstn=None):
        super().__init__(intf, clk=clk, rstn=rstn, allowNoReset=True)
        
        self.requests = []
        self.readPending = False
        self.readed = []

    def doReq(self, s, req):
        rw = req[0]
        addr = req[1]
        
        if rw == READ:
            rw = 0
            wdata = None
            self.readPending = True
            
        elif rw == WRITE:
            wdata = req[2] 
            rw = 1
        else:
            raise NotImplementedError(rw)
        
        intf = self.intf
        s.w(rw, intf.we)
        s.w(addr, intf.addr)
        s.w(wdata, intf.din)
    
    def onReadReq(self, s, addr):
        """
        on readReqRecieved
        """
        raise NotImplementedError()

    def onWriteReq(self, s, addr, data):
        raise NotImplementedError()
    
    def monitor(self, s):
        intf = self.intf
        
        yield s.updateComplete
        if self.enable and s.read(intf.en).val:
            we = self.read(intf.we)
            addr = self.read(intf.addr)
            if we.val:
                data = self.read(intf.din)
                self.onWriteReq(s, addr, data)
            else:
                self.onReadReq(s, addr)

    def getDrivers(self):
        drivers = [self.driver]
        try:
            clk = self.intf.clk
        except AttributeError:
            return drivers
        drivers.append(oscilate(clk))
        return drivers
         
        
    def driver(self, s):
        intf = self.intf
        readPending = self.readPending
        if self.requests and self.enable:
            req = self.requests.pop(0)
            if req is None:
                s.w(0, intf.en)
                self.readPending = False
            else:    
                self.doReq(s, req)
                s.w(1, intf.en)
        else:
            s.w(0, intf.en)
            self.readPending = False
        
        if readPending:
            yield s.updateComplete
            d = s.r(intf.dout)
            self.readed.append(d)
    


class BramPortAgent(BramPort_withoutClkAgent):
    def __init__(self, intf, clk=None, rstn=None):
        if clk is None:
            clk = intf.clk
        super().__init__(intf, clk=clk, rstn=rstn)
        
    def getSubDrivers(self):
        yield oscilate(self.intf.clk)




class VhdlCodeWrap():
    def __init__(self, code_str):
        self.code_str = code_str
        
    def asVhdl(self, serializer, createTmpVarFn):
        return self.code_str
    
    def __str__(self):
        return self.asVhdl(None, None)


from hwt.simulator.agentBase import SyncAgentBase
from hwt.hdlObjects.constants import NOP


class HandshakedAgent(SyncAgentBase):
    """
    Simulation/verification agent for handshaked interface
    @attention: requires clk and rst/rstn signal
      (if you do not have any create simulation wrapper with it)
    """
    def __init__(self, intf, clk=None, rstn=None):
        super().__init__(intf, clk=clk, rstn=rstn)
        self.actualData = NOP
        self.data = []
        # these signals are extracted like this to make
        # agent more configurable
        self._rd = self.getRd()
        self._vld = self.getVld()

    def getRd(self):
        """get "ready" signal"""
        return self.intf.rd

    def getVld(self):
        """get "valid" signal"""
        return self.intf.vld

    def monitor(self, s):
        """
        Collect data from interface
        """
        if s.r(self.rst_n).val and self.enable:
            s.w(1, self._rd)

            yield s.updateComplete
            vld = s.r(self._vld)
            if vld.val or not vld.vldMask:
                d = self.doRead(s)
                if self._debugOutput is not None:
                    self._debugOutput.write("%s, read, %d: %r\n" % (
                                            self.intf._getFullName(), s.now, d))
                self.data.append(d)
        else:
            s.w(0, self._rd)

    def doRead(self, s):
        """extract data from interface"""
        return s.read(self.intf.data)

    def doWrite(self, s, data):
        """write data to interface"""
        s.w(data, self.intf.data)

    def driver(self, s):
        """Push data to interface"""
        if self.actualData is NOP and self.data:
            self.actualData = self.data.pop(0)

        do = self.actualData is not NOP

        if do:
            self.doWrite(s, self.actualData)
        else:
            self.doWrite(s, None)

        if s.r(self.rst_n).val and do and self.enable:
            s.w(1, self._vld)
        else:
            s.w(0, self._vld)
            return

        yield s.updateComplete

        rd = s.r(self._rd)
        if rd.val or not rd.vldMask:
            if self._debugOutput is not None:
                self._debugOutput.write("%s, wrote, %d: %r\n" % (
                                           self.intf._getFullName(), s.now, self.actualData))
            if self.data:
                self.actualData = self.data.pop(0)
            else:
                self.actualData = NOP


class HandshakeSyncAgent(HandshakedAgent):
    """
    Simulation/verification agent for HandshakedSycn interface
    @attention: there is no data channel on this interface it is synchronization only
    """

    def doWrite(self, s, data):
        pass

    def doRead(self, s):
        raise NotImplementedError()
